"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Converts the episode artifact into flattened tables
"""

import copy
import time
import typing
import warnings
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path

import flatten_dict
import gymnasium
import numpy as np
import pandas as pd
from pydantic import BaseModel
from ray.rllib.utils.spaces.repeated import Repeated
from tqdm import tqdm

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.evaluation_outcome import EvaluationOutcome
from corl.evaluation.recording.folder import FolderRecord
from corl.libraries.units import Quantity
from corl.visualization.network_explainability.env_policy_transforms import _to_flattened_dict


# The dataclasses act as schemas for the dataframes
@dataclass
class AgentDoneStatus:
    """A flattened structure for the episode artifact done status information
    (Win, Loss, Draws) for each agent"""

    test_case: int
    agent_id: str
    done_name: str
    done_status: str | None = None
    """A done status code may not exist if it was not triggered"""


@dataclass
class PlatformDoneStates:
    """A flattened structure for the episode artifact done states (e.g. which
    dones are triggered) for each platform"""

    test_case: int
    platform_name: str
    done_name: str
    done_triggered: bool


@dataclass
class EpisodeMetaData:
    """A flattened structure for the episode artifact metadata"""

    test_case: int
    worker_index: int
    wall_time_sec: float
    frame_rate: float | None
    steps: int
    seconds: float


@dataclass
class FlattenedAgentStep:
    """A flattened structure for the agent steps"""

    test_case: int
    step: int
    """step index"""
    step_rel_to_end: int | None
    """step index relative to the end of the trajectory"""
    time: float | int | None
    """time = step / frame_rate"""
    time_rel_to_end: float | int | None
    """time, relative to the end of the trajectory"""
    agent_id: str
    attribute_descrip: str
    attribute_name: str
    attribute_value: typing.Any
    units: str | None = None

    @dataclass
    class AttributeDescription:
        """A string that describes the attribute"""

        observation: str = "observation"
        action: str = "action"
        reward: str = "reward"
        cumulative_reward: str = "cumulative_reward"


@dataclass
class SpaceData:
    """A flattened structure containing space definitions and units"""

    agent_id: str
    attribute_name: str
    space_type: str
    min: int | float  # noqa: A003
    max: int | float  # noqa: A003
    normalized_min: int | float
    normalized_max: int | float
    units: str | None = "dimensionless"


class EpisodeArtifactTablesValidator(BaseModel):
    """Validator for EpisodeArtifactTables"""

    input_path: Path
    file_loader: typing.Callable


class EpisodeArtifactTables:
    """Creates flattened tables (e.g. dataframes) for the attributes within the EpisodeArtifact"""

    # reducer param for flattened_dict
    _reducer = "dot"
    _reducer_map = {"dot": "."}

    def __init__(
        self,
        input_path: str | Path,
        file_loader: typing.Callable = FolderRecord,
        status_bar: typing.Callable = tqdm,
        print_func: typing.Callable = print,
    ) -> None:
        input_path = Path(input_path)
        self.config: EpisodeArtifactTablesValidator = EpisodeArtifactTablesValidator(input_path=input_path, file_loader=file_loader)
        self.reader = self.config.file_loader(input_path)
        self.evaluation_outcome: EvaluationOutcome = self.load_episode_artifacts()
        self.status_bar = status_bar
        self.print_func = print_func

        # Contains the reset states for the simulator
        self.initial_states_table = pd.DataFrame(self.evaluation_outcome.test_cases).reset_index(drop=False)

        self.agent_dones_table = pd.DataFrame([], columns=list(AgentDoneStatus.__dataclass_fields__.keys()))
        self.platform_dones_table = pd.DataFrame([], columns=list(PlatformDoneStates.__dataclass_fields__.keys()))
        self.episode_metadata_table = pd.DataFrame([], columns=list(EpisodeMetaData.__dataclass_fields__.keys()))
        self.agent_steps_table = pd.DataFrame([], columns=list(FlattenedAgentStep.__dataclass_fields__.keys()))
        self.space_data_table = pd.DataFrame([], columns=list(SpaceData.__dataclass_fields__.keys()))

        # Populates the tables defined above
        self._populate_tables()

        self.tables = {
            "InitialStates": self.initial_states_table,
            EpisodeMetaData.__name__: self.episode_metadata_table,
            AgentDoneStatus.__name__: self.agent_dones_table,
            PlatformDoneStates.__name__: self.platform_dones_table,
            FlattenedAgentStep.__name__: self.agent_steps_table,
            SpaceData.__name__: self.space_data_table,
        }

    def load_episode_artifacts(self) -> EvaluationOutcome:
        """Calls the reader to load episode artifact"""
        return self.reader.load()

    def _populate_tables(self):
        episode_metadata = []
        agent_done_statuses = []
        platform_done_states = []
        steps = []

        space_data_table = self._populate_space_data_table()

        for _, epi_artifact in self.status_bar(self.evaluation_outcome.episode_artifacts.items()):
            episode_metadata.append(
                EpisodeMetaData(
                    test_case=epi_artifact.test_case,
                    frame_rate=epi_artifact.frame_rate,
                    worker_index=epi_artifact.worker_index,
                    wall_time_sec=epi_artifact.duration_sec,
                    steps=len(epi_artifact.steps),
                    seconds=1 / epi_artifact.frame_rate * len(epi_artifact.steps),
                )
            )

            # Process Agent Dones
            agent_done_statuses.extend(self._generate_rows_agent_dones(episode_artifact=epi_artifact))

            # Process Platform Dones
            platform_done_states.extend(self._generate_rows_platform_dones(episode_artifact=epi_artifact))

            # Process Agent Steps
            steps.extend(self._generate_rows_agent_steps(episode_artifact=epi_artifact, space_data_table=space_data_table))

        self.print_func("Populating space data table...")
        self.space_data_table = pd.DataFrame(space_data_table)
        self.print_func("Populating agent done status table...")
        self.agent_dones_table = pd.DataFrame(agent_done_statuses)
        self.print_func("Populating platform done states table...")
        self.platform_dones_table = pd.DataFrame(platform_done_states)
        start = time.time()
        self.print_func("Populating steps table...this could take awhile")
        self.agent_steps_table = pd.DataFrame(steps)
        end = time.time()
        self.print_func(f"|----Took {end-start} seconds")
        self.print_func("Populating episode metadata table...")
        self.episode_metadata_table = pd.DataFrame(episode_metadata)

    def _generate_rows_agent_dones(self, episode_artifact: EpisodeArtifact) -> list[AgentDoneStatus]:  # noqa: PLR6301
        """Generates rows for the agent dones"""
        rows = []
        assert episode_artifact.test_case is not None
        for agent_id, done_status_dict in episode_artifact.episode_state.items():
            for done_name, done_status in done_status_dict.items():
                rows.append(
                    AgentDoneStatus(
                        agent_id=agent_id, done_name=done_name, done_status=done_status.name, test_case=episode_artifact.test_case
                    )
                )
        return rows

    def _generate_rows_platform_dones(self, episode_artifact: EpisodeArtifact) -> list[PlatformDoneStates]:  # noqa: PLR6301
        """Generates rows for the platform dones"""
        row = []
        assert episode_artifact.test_case is not None
        for platform_name, platform_done_dict in episode_artifact.dones.items():
            for done_name, done_state in platform_done_dict.items():
                row.append(
                    PlatformDoneStates(
                        platform_name=platform_name, done_name=done_name, done_triggered=done_state, test_case=episode_artifact.test_case
                    )
                )
        return row

    def _generate_rows_agent_steps(self, episode_artifact: EpisodeArtifact, space_data_table: list[SpaceData]) -> list[FlattenedAgentStep]:
        """Generates rows for the agent steps"""
        row: list[FlattenedAgentStep] = []
        assert episode_artifact.test_case is not None
        if episode_artifact.space_definitions is None:
            warnings.warn("Space definitions missing from episode artifact. Cannot process agent steps.")
            return row

        episode_length = len(episode_artifact.steps)
        for idx, step in enumerate(episode_artifact.steps):
            step_idx = idx + 1
            step_idx_reversed = step_idx - episode_length
            step_time_index_kwargs = {
                "step": step_idx,
                "step_rel_to_end": step_idx_reversed,
                "time": step_idx / episode_artifact.frame_rate if episode_artifact.frame_rate else None,
                "time_rel_to_end": step_idx_reversed / episode_artifact.frame_rate if episode_artifact.frame_rate else None,
            }
            for agent_name, agent_step in step.agents.items():
                platform_names = episode_artifact.agent_to_platforms[agent_name]
                # #############################################################
                # Observations: Creates a row of data for each observation
                # #############################################################
                flattened_obs = self._process_raw_observation(agent_step.observations, multiplatform=len(platform_names) > 1)
                obs_space = episode_artifact.space_definitions.observation_space[agent_name]
                # Process multiplatform observation space. This will get rid of the top-level
                # Repeated space.
                if len(platform_names) > 1:
                    assert isinstance(obs_space, gymnasium.spaces.Dict)
                    obs_space = self._restructure_multiplatform_obs_space(obs_space)

                # Perform a simple flattening (this will not flatten the Repeated spaces in the nested dict)
                # NOTE: we do not call _to_flatten_dict here because we want to preserve the Repeated space
                # types so that we can parse the observation values accordingly below.
                flattened_obs_space = flatten_dict.flatten(obs_space, reducer=self._reducer)
                for obs_name, obs_value in flattened_obs.items():
                    # #########################################################
                    # Handles Repeated Observations
                    # #########################################################
                    if isinstance(flattened_obs_space[obs_name], Repeated):
                        # RepeatedObservations are structured as List[Dict[str, Any]]. Iterate through the repeated
                        # observations. Each repeated observation is processed and flattened into the table.
                        # For each element in the top level list
                        for idx, packed_obs in enumerate(obs_value):  # noqa: PLW2901
                            # Each element should be a dictionary, if it is not, then skip and warn.
                            if not isinstance(packed_obs, dict):
                                warnings.warn(f"Skipping: {obs_name}. Expected dict got: {type(packed_obs)}")
                                continue
                            for key, value in packed_obs.items():
                                # TODO: this key is wrong
                                attribute_name_suffix = f"[{idx}]"
                                value = _return_python_typed_value(_get_single_array_element(value))  # noqa: PLW2901
                                # #############################################
                                # Handles multi-dimensional fields
                                # #############################################
                                row = self._create_row_foreach_array_element(
                                    input_value=value,
                                    stored_rows=row,
                                    attribute_name=f"{obs_name}{self._reducer_map[self._reducer]}{key}",
                                    agent_id=agent_name,
                                    attribute_descrip=FlattenedAgentStep.AttributeDescription.observation,
                                    test_case=episode_artifact.test_case,
                                    space_data_table=space_data_table,
                                    attribute_name_suffix=attribute_name_suffix,
                                    **step_time_index_kwargs,
                                )
                    # #########################################################
                    # Non-Repeated Obs
                    # #########################################################
                    else:
                        obs_value = _return_python_typed_value(_get_single_array_element(obs_value))  # noqa: PLW2901
                        # #####################################################
                        # Handles multi-dimensional fields
                        # #####################################################

                        row = self._create_row_foreach_array_element(
                            input_value=obs_value,
                            attribute_name=obs_name,
                            stored_rows=row,
                            agent_id=agent_name,
                            attribute_descrip=FlattenedAgentStep.AttributeDescription.observation,
                            test_case=episode_artifact.test_case,
                            space_data_table=space_data_table,
                            attribute_name_suffix=None,
                            **step_time_index_kwargs,
                        )
                # #############################################################
                # Actions: Creates a row of data for each action
                # #############################################################
                flattened_actions = self._process_raw_actions(agent_step.actions)
                for action_name, action_value in flattened_actions.items():
                    action_value = _return_python_typed_value(_get_single_array_element(action_value))  # noqa: PLW2901
                    row = self._create_row_foreach_array_element(
                        input_value=action_value,
                        attribute_name=action_name,
                        stored_rows=row,
                        agent_id=agent_name,
                        attribute_descrip=FlattenedAgentStep.AttributeDescription.action,
                        test_case=episode_artifact.test_case,
                        space_data_table=space_data_table,
                        attribute_name_suffix=None,
                        **step_time_index_kwargs,
                    )
                # Rewards: Creates a row for each reward
                if agent_step.rewards:
                    for reward_name, rewards_value in agent_step.rewards.items():
                        row.append(
                            FlattenedAgentStep(
                                agent_id=agent_name,
                                attribute_descrip=FlattenedAgentStep.AttributeDescription.reward,
                                attribute_name=reward_name,
                                attribute_value=rewards_value,
                                test_case=episode_artifact.test_case,
                                step=step_idx,
                                step_rel_to_end=step_idx_reversed,
                                time=step_time_index_kwargs["time"],
                                time_rel_to_end=step_time_index_kwargs["time_rel_to_end"],
                            )
                        )
                # Total Reward: Creates one row for the cumulative reward (total reward)
                row.append(
                    FlattenedAgentStep(
                        agent_id=agent_name,
                        attribute_descrip=FlattenedAgentStep.AttributeDescription.cumulative_reward,
                        attribute_name="total_reward",
                        attribute_value=agent_step.total_reward,
                        test_case=episode_artifact.test_case,
                        step=step_idx,
                        step_rel_to_end=step_idx_reversed,
                        time=step_time_index_kwargs["time"],
                        time_rel_to_end=step_time_index_kwargs["time_rel_to_end"],
                    )
                )
        return row

    def _populate_space_data_table(self):
        """Populates the space data table"""
        # The space definitions and units are the same across all episode
        # artifacts, so we only need to process one.
        rows = []

        epi_artifact = self.evaluation_outcome.episode_artifacts[0]
        space_definitions = epi_artifact.space_definitions

        if space_definitions is not None:
            flattened_obs_units = {}
            for agent_id, _units in epi_artifact.observation_units.items():
                flattened_obs_units[agent_id] = _to_flattened_dict(_units, reducer=self._reducer)

            rows.extend(
                self._generate_rows_space_data_table(
                    space_dict=space_definitions.action_space.spaces,
                    normalized_space_dict=space_definitions.normalized_action_space.spaces,
                    obs_units=flattened_obs_units,
                )
            )

            rows.extend(
                self._generate_rows_space_data_table(
                    space_dict=space_definitions.observation_space.spaces,
                    normalized_space_dict=space_definitions.normalized_observation_space.spaces,
                    obs_units=flattened_obs_units,
                )
            )

        return rows

    def _generate_rows_space_data_table(self, space_dict: OrderedDict, normalized_space_dict: OrderedDict, obs_units: dict):
        """Generates rows for the space data table given the spaces dict and attribute description"""
        rows: list = []
        # Iterate over the agents
        for agent_id, space_def in space_dict.items():
            flattened_space_def = _to_flattened_dict(space_def, reducer=self._reducer)
            flattened_normalized_space_def = _to_flattened_dict(normalized_space_dict[agent_id], reducer=self._reducer)
            obs_unit_one_agent = obs_units.get(agent_id, {})
            # Iterate over the spaces for each agent
            # The space_def comes from the original space. For the observation space, this may contain the sensors that were
            # excluded from training. The normalized_space_dict filters these sensors out.
            for flattened_key, normalized_space in flattened_normalized_space_def.items():
                assert isinstance(normalized_space, gymnasium.Space)
                min_val, max_val = get_min_max_from_gymnasium_space(flattened_space_def[flattened_key])
                min_val_normed, max_val_normed = get_min_max_from_gymnasium_space(normalized_space)
                # Explicit typing for mypy
                if isinstance(min_val, np.ndarray) and isinstance(max_val, np.ndarray):
                    assert isinstance(max_val_normed, np.ndarray) and isinstance(min_val_normed, np.ndarray)
                    rows = self._process_array_attribute_space_data(
                        input_min=min_val,
                        input_max=max_val,
                        input_min_normed=min_val_normed,
                        input_max_normed=max_val_normed,
                        stored_rows=rows,
                        attribute_name=flattened_key,
                        agent_id=agent_id,
                        space_type=normalized_space.__class__.__name__,
                        units=obs_unit_one_agent.get(flattened_key),
                    )
                if isinstance(min_val, int | float) and isinstance(max_val, int | float):
                    assert isinstance(min_val_normed, int | float) and isinstance(max_val_normed, int | float)
                    _units = _get_single_array_element(obs_unit_one_agent.get(flattened_key))
                    rows.append(
                        SpaceData(
                            agent_id=agent_id,
                            attribute_name=flattened_key,
                            space_type=normalized_space.__class__.__name__,
                            min=min_val,
                            max=max_val,
                            normalized_min=min_val_normed,
                            normalized_max=max_val_normed,
                            units=_units,
                        )
                    )
        return rows

    @staticmethod
    def _process_array_attribute_space_data(
        input_min: np.ndarray,
        input_max: np.ndarray,
        input_min_normed: np.ndarray,
        input_max_normed: np.ndarray,
        stored_rows: list[SpaceData],
        attribute_name,
        units: list | (None | str),
        **space_data_kwargs,
    ) -> list[SpaceData]:
        """Helper function to flatten various array fields in the space data table."""
        input_min = input_min.flatten()
        input_max = input_max.flatten()
        input_min_normed = input_min_normed.flatten()
        input_max_normed = input_max_normed.flatten()

        assert len(input_min) == len(input_max)
        n_elements = len(input_min)
        for idx in range(n_elements):
            min_val = _return_python_typed_value(input_min[idx])
            max_val = _return_python_typed_value(input_max[idx])
            min_val_norm = _return_python_typed_value(input_min_normed[idx])
            max_val_norm = _return_python_typed_value(input_max_normed[idx])
            # Creates a zero-padded index for the attribute
            attribute_padded_idx = str(idx).zfill(len(str(n_elements)))
            new_attribute_name = f"{attribute_name}_{attribute_padded_idx }"
            if isinstance(units, list | tuple):
                _unit = units[idx]
            elif isinstance(units, str):
                _unit = units
            elif units is None:
                _unit = None
            else:
                raise ValueError(f"Units is an unknown type: units: {units}, type: {type(units)}")
            stored_rows.append(
                SpaceData(
                    min=min_val,
                    max=max_val,
                    normalized_min=min_val_norm,
                    normalized_max=max_val_norm,
                    attribute_name=new_attribute_name,
                    units=_unit,
                    **space_data_kwargs,
                )
            )
        return stored_rows

    @staticmethod
    def _create_row_foreach_array_element(
        input_value: tuple | (list | (np.ndarray | (float | int))),
        stored_rows: list[FlattenedAgentStep],
        attribute_name: str,
        space_data_table: list[SpaceData] | None = None,
        attribute_name_suffix: str | None = None,
        **flatten_agent_step_kwargs,
    ) -> list[FlattenedAgentStep]:
        """
        Helper function to process the array attributes. Creates a new column
        for each element in the input, where the new attribute name contains an index
        corresponding to the element of its original array. When the input value is not an array,
        a single row is added.
        """
        # #####################################################################
        # Handles multi-dimensional fields
        # #####################################################################
        if isinstance(input_value, np.ndarray | list | tuple):
            if isinstance(input_value, np.ndarray):
                input_value = input_value.flatten()

            n_elements = len(input_value)
            for _attribute_idx, _value in enumerate(input_value):
                _value = _return_python_typed_value(_value)
                # zero pads up to the max number of elements - this makes sorting the attribute names easier
                attribute_idx_str = str(_attribute_idx).zfill(len(str(n_elements)))
                # This is the structure of the attribute name in the space table for arrays
                new_attribute_name = f"{attribute_name}_{attribute_idx_str}"
                if space_data_table is not None:
                    units = None
                    units = [
                        item.units
                        for item in space_data_table
                        if (item.agent_id == flatten_agent_step_kwargs["agent_id"]) and (item.attribute_name == new_attribute_name)
                    ]
                    assert len(units) <= 1
                    unit = units[0] if len(units) > 0 else None
                # When a suffix is passed in (this is for repeated obs), it will go before the array index
                if attribute_name_suffix:
                    new_attribute_name = f"{attribute_name}{attribute_name_suffix}_{attribute_idx_str}"
                stored_rows.append(
                    FlattenedAgentStep(attribute_name=new_attribute_name, attribute_value=_value, units=unit, **flatten_agent_step_kwargs)
                )
            return stored_rows
        # #####################################################################
        # Handle single-valued fields
        # #####################################################################
        if space_data_table is not None:
            units = None
            units = [
                item.units
                for item in space_data_table
                if (item.agent_id == flatten_agent_step_kwargs["agent_id"]) and (item.attribute_name == attribute_name)
            ]
            assert len(units) <= 1
            unit = units[0] if len(units) > 0 else None
        if attribute_name_suffix:
            attribute_name = f"{attribute_name}{attribute_name_suffix}"
        stored_rows.append(
            FlattenedAgentStep(attribute_name=attribute_name, attribute_value=input_value, units=unit, **flatten_agent_step_kwargs)
        )
        return stored_rows

    def _restucture_multiplatform_obs(self, observations: OrderedDict) -> OrderedDict:  # noqa: PLR6301
        """Restructures the multiplatform observaton.
        The multiplatform observations have the structure: OrderedDict(platform_name, [OrderedDict(obs_name, ...)])"""
        observations = copy.deepcopy(observations)
        restructured_obs = OrderedDict()
        error_msg = "The assumption of how the multiplatform observations are structured is incorrect"
        for _platform_name, obs in observations.items():
            assert len(obs) == 1, error_msg
            restructured_obs[_platform_name] = obs[0]
        return restructured_obs

    def _restructure_multiplatform_obs_space(self, obs_space: gymnasium.spaces.Dict) -> gymnasium.spaces.Dict:  # noqa: PLR6301
        """Restructures the multiplatform observation space. The multiplatform observation space is structured as
        {agent name: {platform name: Repeated(Dict(...), ..})}}. This removes
        the top-level Repeated space, so it flattens appropriately"""
        obs_space = copy.deepcopy(obs_space)
        restructured_space = gymnasium.spaces.Dict()
        for _platform_name, _space in obs_space.items():
            assert isinstance(_space, Repeated), f"Expected Repeated Space, got: {type(_space)}"
            restructured_space[_platform_name] = _space.child_space
        return restructured_space

    def _process_raw_observation(self, observation: OrderedDict, multiplatform: bool = False):
        """Processes the raw observation"""
        tmp = observation.copy()
        for idx, val in observation.items():
            for key, obs in val.items():
                if isinstance(obs, Quantity):
                    tmp[idx][key] = obs.m
        observation = tmp

        if multiplatform:
            return flatten_dict.flatten(self._restucture_multiplatform_obs(observation), reducer=self._reducer)
        return flatten_dict.flatten(observation, reducer=self._reducer)

    def _process_raw_actions(self, action: OrderedDict):
        return flatten_dict.flatten(action, reducer=self._reducer)


def _get_single_array_element(input_value: typing.Any):
    """When an iterable contains a single element, returns the element"""
    if isinstance(input_value, np.ndarray):
        if input_value.shape and max(input_value.shape) == 1:
            return input_value[0]
        # In some cases, values are stored as "no length arrays" where the type is an array
        # but it has no shape, when this is the case, calling .tolist() on it returns the underlying
        # value in the array
        if len(input_value.shape) == 0:
            return input_value.tolist()

    if isinstance(input_value, list | tuple) and len(input_value) == 1:
        return input_value[0]

    return input_value


def _return_python_typed_value(input_value: typing.Any):
    """When the input value is a numpy generic type (np.float32, etc.)
    returns the python equivalent"""
    if isinstance(input_value, np.generic):
        return input_value.item()
    return input_value


def get_min_max_from_gymnasium_space(space: gymnasium.Space):
    """Returns the minimum and maximum of a gymnasium space"""
    if isinstance(space, gymnasium.spaces.Discrete):
        min_val = space.start
        max_val = min_val + space.n
        return _return_python_typed_value(min_val), _return_python_typed_value(max_val)

    if isinstance(space, gymnasium.spaces.MultiDiscrete):
        max_val = _get_single_array_element(space.nvec)
        min_val = _get_single_array_element(np.zeros_like(max_val))
        return _return_python_typed_value(min_val), _return_python_typed_value(max_val)

    if isinstance(space, gymnasium.spaces.MultiBinary):
        min_val = _get_single_array_element(np.zeros(shape=space.shape))
        max_val = _get_single_array_element(min_val + 1)
        return _return_python_typed_value(min_val), _return_python_typed_value(max_val)

    if isinstance(space, gymnasium.spaces.Box):
        min_val = _get_single_array_element(space.low)
        max_val = _get_single_array_element(space.high)
        return _return_python_typed_value(min_val), _return_python_typed_value(max_val)

    raise NotImplementedError(f"An unsupported space was provided: {space}")



"""
---------------------------------------------------------------------------.

Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import os
import pathlib
import time
import typing

import ray
from pydantic import ConfigDict, ImportString, field_validator
from pyinstrument import Profiler
from ray.tune.registry import get_trainable_cls

from corl.environment.default_env_rllib_callbacks import EnvironmentDefaultCallbacks
from corl.environment.multi_agent_env import ACT3MultiAgentEnv, ACT3MultiAgentEnvValidator
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.episode_parameter_providers.remote import RemoteEpisodeParameterProvider
from corl.experiments.base_experiment import BaseExperiment, BaseExperimentValidator, ExperimentFileParse
from corl.libraries.factory import Factory
from corl.parsers.yaml_loader import apply_patches
from corl.policies.base_policy import BasePolicyValidator


class BenchmarkExperimentValidator(BaseExperimentValidator):
    """
    ray_config: dictionary to be fed into ray init, validated by ray init call
    env_config: environment configuration, validated by environment class
    rllib_configs: a dictionary
    Arguments:
        BaseModel {[type]} -- [description].

    Raises
    ------
        RuntimeError: [description]

    Returns
    -------
        [type] -- [description]
    """

    ray_config: dict[str, typing.Any]
    env_config: dict[str, typing.Any]
    rllib_configs: dict[str, dict[str, typing.Any]]
    tune_config: dict[str, typing.Any]
    trainable_config: dict[str, typing.Any] | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)
    benchmark_episodes: int = 10

    @field_validator("rllib_configs", mode="before")
    @classmethod
    def apply_patches_rllib_configs(cls, v):
        """
        The dictionary of rllib configs may come in as a dictionary of
        lists of dictionaries, this function is responsible for collapsing
        the list down to a typing.Dict[str, typing.Dict[str, typing.Any]]
        instead of
        typing.Dict[str, typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]].

        Raises
        ------
        RuntimeError: [description]

        Returns
        -------
        [type] -- [description]
        """
        if not isinstance(v, dict):
            raise TypeError("rllib_configs are expected to be a dict of keys to different compute configs")
        rllib_configs = {}
        for key, value in v.items():
            if isinstance(value, list):
                rllib_configs[key] = apply_patches(value)
            elif isinstance(value, dict):
                rllib_configs[key] = value
        return rllib_configs

    @field_validator("ray_config", "tune_config", "trainable_config", "env_config", mode="before")
    @classmethod
    def apply_patches_configs(cls, v):
        """
        reduces a field from
        typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]
        to
        typing.Dict[str, typing.Any].

        by patching the first dictionary in the list with each patch afterwards

        Returns
        -------
        [type] -- [description]
        """
        if isinstance(v, list):
            v = apply_patches(v)
        return v


class RllibPolicyValidator(BasePolicyValidator):
    """
    policy_class: callable policy class None will use default from trainer
    train: should this policy be trained
    Arguments:
        BaseModel {[type]} -- [description].

    Raises
    ------
        RuntimeError: [description]

    Returns
    -------
        [type] -- [description]
    """

    config: dict[str, typing.Any] = {}
    policy_class: ImportString | None = None
    train: bool = True


class BenchmarkExperiment(BaseExperiment):
    """
    The Rllib Experiment is an experiment for running
    multiagent configurable environments with patchable settings.
    """

    def __init__(self, **kwargs) -> None:
        self.config: BenchmarkExperimentValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BenchmarkExperimentValidator]:
        return BenchmarkExperimentValidator

    @property
    def get_policy_validator(self) -> type[RllibPolicyValidator]:
        """Return validator."""
        return RllibPolicyValidator

    def run_experiment(self, args: ExperimentFileParse) -> float:
        rllib_config = self._select_rllib_config(args.compute_platform)
        if args.compute_platform in ["ray"]:
            self._update_ray_config_for_ray_platform()

        if args.debug:
            rllib_config["num_workers"] = 0
            self.config.ray_config["local_mode"] = True

        self._add_trial_creator()

        ray.init(**self.config.ray_config)

        self.config.env_config["agents"], self.config.env_config["agent_platforms"] = self.create_agents(
            args.platform_config,
            args.agent_config,
        )

        # self.config.env_config["horizon"] = rllib_config["horizon"]

        if not self.config.ray_config["local_mode"]:
            self.config.env_config["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                Factory(**self.config.env_config["episode_parameter_provider"]),
                actor_name=ACT3MultiAgentEnv.episode_parameter_provider_name,
            )

            for agent_name, agent_configs in self.config.env_config["agents"].items():
                agent_configs.class_config.config["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                    Factory(**agent_configs.class_config.config["episode_parameter_provider"]),
                    agent_name,
                )

        self.config.env_config["epp_registry"] = ACT3MultiAgentEnvValidator(**self.config.env_config).epp_registry

        tmp = ACT3MultiAgentEnv(self.config.env_config)
        act_space = tmp.action_space

        env = tmp

        def save_episode_information(_, __, ___, ____):
            pass

        env._simulator.save_episode_information = save_episode_information  # type: ignore  # noqa: SLF001
        env.config.deep_sanity_check = True
        profiler = Profiler()
        profiler.start()

        # retrieve action
        # if sanity_check_state_dict:
        total_timesteps = 0
        t1 = time.time()

        for _ep in range(self.config.benchmark_episodes):
            print(f"Episode {_ep}")
            time.time()
            env.reset()

            # if debug_print:
            done = False
            step = 0

            while not done:
                multi_actions = self.generate_action(act_space)
                _, _, trainable_done, _, _ = env.step(multi_actions)
                done = trainable_done["__all__"]
                step += 1
            total_timesteps += step
            print(f"\tSteps = {step}")
            time.time()
        t2 = time.time()

        profiler.stop()

        step_per_second = total_timesteps / (t2 - t1)
        print(f"Total Steps  {total_timesteps}")
        print(f"Steps/Second {step_per_second}")

        open_file_name = "testing.html"
        if hasattr(self, "experiment_config_id"):
            open_file_name = f"testing_{self.experiment_config_id}_{int(round(step_per_second))}.html"

        with open(open_file_name, "w", encoding="utf-8") as f_html:
            f_html.write(profiler.output_html())

        return step_per_second

    def generate_action(self, act_space):  # noqa: PLR6301
        """randomly select an action to take."""
        # generate a random action
        return {a_k: {s_k: s.sample() for s_k, s in a_s.spaces.items()} for a_k, a_s in act_space.spaces.items()}

    def get_callbacks(self) -> type[EnvironmentDefaultCallbacks]:  # noqa: PLR6301
        """Get the environment callbacks."""
        return EnvironmentDefaultCallbacks

    def _select_rllib_config(self, platform: str | None) -> dict[str, typing.Any]:
        """Extract the rllib config for the proper computational platform.

        Parameters
        ----------
        platform : typing.Optional[str]
            Specification of the computational platform to use, such as "local", "hpc", etc.  This must be present in the rllib_configs.
            If None, the rllib_configs must only have a single entry.

        Returns
        -------
        typing.Dict[str, typing.Any]
            Rllib configuration for the desired computational platform.

        Raises
        ------
        RuntimeError
            The requested computational platform does not exist or None was used when multiple platforms were defined.
        """
        if platform is not None:
            return self.config.rllib_configs[platform]

        if len(self.config.rllib_configs) == 1:
            return self.config.rllib_configs[next(iter(self.config.rllib_configs))]

        raise RuntimeError(f'Invalid rllib_config for platform "{platform}"')

    def _update_ray_config_for_ray_platform(self) -> None:
        """Update the ray configuration for ray platforms."""
        self.config.ray_config["address"] = "auto"
        self.config.ray_config["log_to_driver"] = False

    def _enable_episode_parameter_provider_checkpointing(self) -> None:
        base_trainer = self.config.tune_config["run_or_experiment"]

        trainer_class = get_trainable_cls(base_trainer)

        class EpisodeParameterProviderSavingTrainer(trainer_class):  # type: ignore[valid-type, misc]
            """
            Tune Trainer that adds capability to restore
            progress of the EpisodeParameterProvider on restoring training
            progress.
            """

            def save_checkpoint(self, checkpoint_path: str) -> None:
                """
                adds additional checkpoint saving functionality
                by also saving any episode parameter providers
                currently running.
                """
                tmp = super().save_checkpoint(checkpoint_path)

                checkpoint_folder = pathlib.Path(checkpoint_path)

                # Environment
                epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
                env = self.workers.local_worker().env
                epp: EpisodeParameterProvider = env.config.epp
                epp.save_checkpoint(checkpoint_folder / epp_name)

                # Agents
                for agent_name, agent_configs in env.agent_dict.items():
                    epp = agent_configs.config.epp
                    epp.save_checkpoint(checkpoint_folder / agent_name)

                return tmp

            def load_checkpoint(self, checkpoint_path: str) -> None:
                """
                adds additional checkpoint loading functionality
                by also loading any episode parameter providers
                with the checkpoint.
                """
                super().load_checkpoint(checkpoint_path)

                checkpoint_folder = pathlib.Path(checkpoint_path).parent

                # Environment
                epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
                env = self.workers.local_worker().env
                epp: EpisodeParameterProvider = env.config.epp
                epp.load_checkpoint(checkpoint_folder / epp_name)

                # Agents
                for agent_name, agent_configs in env.agent_dict.items():
                    epp = agent_configs.config.epp
                    epp.load_checkpoint(checkpoint_folder / agent_name)

        self.config.tune_config["run_or_experiment"] = EpisodeParameterProviderSavingTrainer

    def _add_trial_creator(self):
        """Updates the trial name based on the HPC Job Number and the trial name in the configuration."""
        if "trial_name_creator" not in self.config.tune_config:

            def trial_name_prefix(trial):
                """
                Args:
                    trial (Trial): A generated trial object.

                Returns
                -------
                trial_name (str): String representation of Trial prefixed
                    by the contents of the environment variable:
                    TRIAL_NAME_PREFIX
                    Or the prefix 'RUN' if none is set.
                """
                trial_prefix = os.environ.get("PBS_JOBID", os.environ.get("TRIAL_NAME_PREFIX", ""))
                trial_name = ""

                if "TrialName" in self.config.env_config:
                    trial_name = "-" + self.config.env_config["TrialName"] if trial_prefix else self.config.env_config["TrialName"]
                return f"{trial_prefix}{trial_name}-{trial}"

            self.config.tune_config["trial_name_creator"] = trial_name_prefix



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import gc
import logging
import traceback
import typing
from collections.abc import Iterator
from threading import Condition
from weakref import CallableProxyType

from pydantic import BaseModel
from ray.rllib.env.base_env import BaseEnv, convert_to_base_env
from ray.rllib.evaluation.env_runner_v2 import EnvRunnerV2, _build_multi_agent_batch, _PerfStats
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.evaluation.rollout_worker import RolloutWorker
from ray.rllib.policy.sample_batch import MultiAgentBatch, SampleBatch, concat_samples
from ray.rllib.utils.typing import EnvID, SampleBatchType

from corl.dones.done_func_base import DoneStatusCodes
from corl.environment.multi_agent_env import ACT3MultiAgentEnv
from corl.evaluation.runners.section_factories.engine.rllib.interruptable_callback import InterruptableCallback


class AlgorithmRunnerValidator(BaseModel):
    """Validator for AlgorithmRunner"""

    exit_on_end: bool = False


class AlgorithmRunner:
    """Runs an environment"""

    def __init__(self, **kwargs):
        """"""
        self.config = self.get_validator()(**kwargs)

        self._algorithm: CallableProxyType | None = None  # type: ignore
        self._base_env: BaseEnv | None = None  # type: ignore
        self._running: bool = True  # type: ignore
        self.__runner: EnvRunnerV2 | None = None  # type: ignore

        self._logger = logging.getLogger(AlgorithmRunner.__name__)

        self._algorithm_cv = Condition()

    @staticmethod
    def get_validator() -> type[AlgorithmRunnerValidator]:
        """Return the validator for AlgorithmRunner"""
        return AlgorithmRunnerValidator

    @property
    def algorithm(self) -> CallableProxyType:
        if self._algorithm is None:
            raise RuntimeError("algorithm has not been initialized")
        return self._algorithm

    @property
    def base_env(self) -> BaseEnv:
        return convert_to_base_env(env=self.local_worker.env)

    @property
    def local_worker(self) -> RolloutWorker:
        if self.algorithm.workers is None:
            raise RuntimeError("No Algorithm workers")

        return self.algorithm.workers.local_worker()

    def reset(self, algorithm: CallableProxyType | None = None):
        """Stops any existing episodes and sets the algorithm used by this _runner"""

        with self._algorithm_cv:
            self._logger.info(f'{"*" * 40}    RESETTING    {"*" * 40}')
            if algorithm is not None:
                algorithm = typing.cast(CallableProxyType, algorithm)  # type: ignore

                self.__runner = None

                # if self._algorithm is not None:
                #     # HACK: Cleanup simulator simulator
                #     def shutdown_env(env):
                #         try:
                #             env.simulator.shutdown()
                #         except (RuntimeError, KeyError, AttributeError):
                #             pass

                #     self._algorithm.workers.foreach_env(shutdown_env)
                #     self._algorithm.stop()
                #     del self._algorithm
                self._algorithm = algorithm

            gc.collect()
            self._algorithm_cv.notify_all()

    @property
    def _runner(self) -> EnvRunnerV2:
        if self.__runner is None:
            multiple_episodes_in_batch = self.algorithm.config.batch_mode != "complete_episodes"
            rollout_fragment_length = int(self.algorithm.config.rollout_fragment_length)
            self.__runner = EnvRunnerV2(
                self.local_worker,
                convert_to_base_env(env=self.local_worker.env),
                multiple_episodes_in_batch=multiple_episodes_in_batch,
                callbacks=self.algorithm.callbacks,
                perf_stats=_PerfStats(),
                rollout_fragment_length=rollout_fragment_length,
            )
        return self.__runner

    def _handle_errored_episode(self, env_id: EnvID, error: Exception) -> Iterator[SampleBatchType]:
        outputs: list[SampleBatch | MultiAgentBatch] = []
        episode: EpisodeV2 = self._runner._active_episodes[env_id]  # noqa: SLF001
        batch_builder = self._runner._batch_builders[env_id]  # noqa: SLF001

        env = self.local_worker.env
        if isinstance(env, ACT3MultiAgentEnv):
            metadata = {}
            metadata.update(env.config.metadata)
            metadata["error"] = str(traceback.format_exc())

            for platform_name in env.state.episode_state:
                env._state.episode_state[platform_name][f"{type(error).__name__}"] = DoneStatusCodes.DRAW  # noqa: SLF001

            env.simulator.mark_episode_done(env._done_info, env._state.episode_state, metadata)  # noqa: SLF001

        episode.postprocess_episode(
            batch_builder=batch_builder,
            is_done=True,
            check_dones=False,
        )

        ma_sample_batch = _build_multi_agent_batch(
            episode.episode_id,
            batch_builder,
            self._runner._large_batch_threshold,  # noqa: SLF001
            self._runner._multiple_episodes_in_batch,  # noqa: SLF001
        )
        if ma_sample_batch:
            outputs.append(ma_sample_batch)

        # SampleBatch built from data collected by batch_builder.
        # Clean up and delete the batch_builder.
        del self._runner._batch_builders[env_id]  # noqa: SLF001

        yield from outputs

    def run(self) -> Iterator[SampleBatchType]:
        def process_outputs(outputs):
            if isinstance(
                outputs,
                SampleBatch | MultiAgentBatch,
            ):
                output = concat_samples([outputs])

                self.local_worker.callbacks.on_sample_end(worker=self.local_worker, samples=output)

                return output
            return None

        while self._running:
            with self._algorithm_cv:
                try:
                    self._logger.debug("Running...")
                    for outputs in self._runner.run():
                        if output := process_outputs(outputs):
                            yield output
                            break

                except Exception as exception:  # noqa: BLE001
                    self._logger.error(exception)
                    for outputs in self._handle_errored_episode(env_id=0, error=exception):
                        if output := process_outputs(outputs):
                            yield output
                finally:
                    self.__runner = None

                    if self.config.exit_on_end:
                        self._running = False

    def stop(self, continue_running=False):
        try:
            for callback in self.algorithm.callbacks._callback_list:  # noqa: SLF001
                if isinstance(callback, InterruptableCallback):
                    callback.stop()

            if not continue_running:
                self._running = False
                self.local_worker.stop()
        except (ReferenceError, RuntimeError):
            pass



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Reward for GymnasiumSimulator
"""
from corl.rewards.reward_func_base import RewardFuncBase


class GymnasiumReward(RewardFuncBase):
    """
    Reward for GymnasiumSimulator that rewards the reward
    coming from the simulator provided state and reports it
    """

    def __call__(
        self,
        observation,
        action,
        next_observation,
        state,
        next_state,
        observation_space,
        observation_units,
    ):
        return next_state.rewards[self.config.platform_names[0]]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import logging
from abc import abstractmethod
from typing import Annotated, Generic, TypeVar

from pydantic import BaseModel, BeforeValidator, ConfigDict

from corl.evaluation.connection.signal import Signal
from corl.evaluation.launchers.base_eval import EvalConfig
from corl.libraries.factory import Factory

EppUpdateT = TypeVar("EppUpdateT")


class BaseEvalConnectionValidator(BaseModel):
    ...


class BaseEvalConnection(Generic[EppUpdateT]):
    def __init__(self, **kwargs):
        self._config = self.get_validator()(**kwargs)
        self._logger = logging.getLogger(type(self).__name__)

        self.reset_signal: Signal[EvalConfig] = Signal[EvalConfig]()  # type: ignore
        self.modify_epp_signal: Signal[EppUpdateT] = Signal[EppUpdateT]()  # type: ignore

    @staticmethod
    def get_validator() -> type[BaseEvalConnectionValidator]:
        return BaseEvalConnectionValidator

    @abstractmethod
    def stop(self):
        """Stops the connection"""


class NullConnection(BaseEvalConnection):
    def stop(self):
        pass


class ConnectionValidator(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    connection: Annotated[BaseEvalConnection, BeforeValidator(Factory.resolve_factory)] = NullConnection()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from abc import abstractmethod

from corl.libraries.units import Quantity


class BaseEppUpdate:
    @abstractmethod
    def get_updates(self) -> dict[tuple[str, ...], Quantity | float | str]:
        ...


def get_updates(updates: BaseEppUpdate, prefix: str | None = None) -> dict[tuple[str, ...], Quantity | float | str]:
    modified_updates: dict[tuple[str, ...], Quantity | float | str] = {}

    key: tuple[str, ...]
    value: Quantity | float | str
    for key, value in updates.get_updates().items():
        if prefix is None:
            modified_updates[key] = value
        elif prefix and key[0] == prefix:
            modified_updates[key[1:]] = value

    return modified_updates



"""
---------------------------------------------------------------------------.

Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import json
import re
import typing
from collections import OrderedDict

import flatten_dict
import pandas as pd
from gymnasium.spaces.dict import Dict
from numpy import ndarray
from pydantic import BaseModel, ConfigDict, ImportString, field_validator
from ray.rllib.utils.spaces.repeated import Repeated

from corl.environment.multi_agent_env import ACT3MultiAgentEnv
from corl.experiments.base_experiment import BaseExperiment, BaseExperimentValidator, ExperimentFileParse
from corl.libraries.env_space_util import SingleLayerDict
from corl.libraries.units import Quantity
from corl.parsers.yaml_loader import apply_patches
from corl.policies.base_policy import BasePolicyValidator


class QuantityEncoder(json.JSONEncoder):
    """Basic encoder of quantity variables

    Args:
        json (_type_): _description_
    """

    def default(self, obj):
        if isinstance(obj, Quantity):
            if isinstance(obj.m, ndarray):
                return obj.m.tolist(), str(obj.unit)
            return obj.m, str(obj.unit)
        if isinstance(obj, BaseModel):
            return dict(obj)
        return json.JSONEncoder.default(self, obj)


class ExportSetupExperimentValidator(BaseExperimentValidator):
    """
    ray_config: dictionary to be fed into ray init, validated by ray init call
    env_config: environment configuration, validated by environment class
    rllib_configs: a dictionary
    Arguments:
        BaseModel {[type]} -- [description].

    Raises
    ------
        RuntimeError: [description]

    Returns
    -------
        [type] -- [description]
    """

    ray_config: dict[str, typing.Any]
    env_config: dict[str, typing.Any]
    rllib_configs: dict[str, dict[str, typing.Any]]
    tune_config: dict[str, typing.Any]
    trainable_config: dict[str, typing.Any] | None = None
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("rllib_configs", mode="before")
    @classmethod
    def apply_patches_rllib_configs(cls, v):
        """
        The dictionary of rllib configs may come in as a dictionary of
        lists of dictionaries, this function is responsible for collapsing
        the list down to a typing.Dict[str, typing.Dict[str, typing.Any]]
        instead of
        typing.Dict[str, typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]].

        Raises
        ------
        RuntimeError: [description]

        Returns
        -------
        [type] -- [description]
        """
        if not isinstance(v, dict):
            raise TypeError("rllib_configs are expected to be a dict of keys to different compute configs")
        rllib_configs = {}
        for key, value in v.items():
            if isinstance(value, list):
                rllib_configs[key] = apply_patches(value)
            elif isinstance(value, dict):
                rllib_configs[key] = value
        return rllib_configs

    @field_validator("ray_config", "tune_config", "trainable_config", "env_config", mode="before")
    @classmethod
    def apply_patches_configs(cls, v):
        """
        reduces a field from
        typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]
        to
        typing.Dict[str, typing.Any].

        by patching the first dictionary in the list with each patch afterwards

        Returns
        -------
        [type] -- [description]
        """
        if isinstance(v, list):
            v = apply_patches(v)
        return v


class RllibPolicyValidator(BasePolicyValidator):
    """
    policy_class: callable policy class None will use default from trainer
    train: should this policy be trained
    Arguments:
        BaseModel {[type]} -- [description].

    Raises
    ------
        RuntimeError: [description]

    Returns
    -------
        [type] -- [description]
    """

    config: dict[str, typing.Any] = {}
    policy_class: ImportString | None = None
    train: bool = True


class ExportSetupExperiment(BaseExperiment):
    """
    The Rllib Experiment is an experiment for running
    multi agent configurable environments with patchable settings.
    """

    def __init__(self, **kwargs) -> None:
        self.config: ExportSetupExperimentValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[ExportSetupExperimentValidator]:
        return ExportSetupExperimentValidator

    @property
    def get_policy_validator(self) -> type[RllibPolicyValidator]:
        """Return validator."""
        return RllibPolicyValidator

    def run_experiment(self, args: ExperimentFileParse) -> None:  # noqa: PLR0915
        self.setup_configs(args)

        tmp: ACT3MultiAgentEnv = ACT3MultiAgentEnv(self.config.env_config)
        tmp.reset()
        tmp.step(tmp.action_space.sample())

        def remove_duplicates(in_str) -> str:
            """Remove duplicates"""
            # split input string separated by space
            in_str_list = in_str.split(" ")

            # now create dictionary using counter method
            # which will have strings as key and their
            # frequencies as value
            UniqW = list(OrderedDict.fromkeys(in_str_list))

            return " ".join(UniqW)

        def filter_string(k) -> str:
            """__summary_"""
            temp_string = k.replace("_", " ")
            temp_string = temp_string.replace(".", " . ")
            temp_string = re.sub(r"(\w)([A-Z])", r"\1 \2", temp_string)
            temp_string = temp_string.title()
            temp_string = " ".join(re.findall("[A-Z][^A-Z]*", temp_string)).title()
            temp_string = re.sub(" +", " ", temp_string)
            temp_string = remove_duplicates(temp_string)
            temp_string = (
                temp_string.replace(" ", "").replace("ObsSensor", "").replace("ObserveSensorRepeated", "").replace("DirectObservation", "")
            )
            return temp_string[:-1] if temp_string[-1] == "." else temp_string

        def get_dataframe(in_space):
            if issubclass(type(in_space), dict | Dict | SingleLayerDict):
                temp_space = flatten_dict.flatten(in_space, reducer="dot")
            else:
                temp_space = {"control": in_space}

            temp_new_space = OrderedDict()
            for k, v in temp_space.items():
                if isinstance(v, Repeated):
                    rep_space = flatten_dict.flatten(v.child_space, reducer="dot")
                    for o_k, o_v in rep_space.items():
                        temp_new_space[f"{k}.{o_k}"] = o_v
                elif isinstance(k, tuple):
                    temp_new_space[".".join(k)] = v
                else:
                    temp_new_space[k] = v

            filtered_temp_space = {
                filter_string(flattened_key): str(flatten_value) for flattened_key, flatten_value in temp_new_space.items()
            }
            return pd.DataFrame(data=filtered_temp_space, index=[0])

        def process_space(in_space, in_space_str, in_space_units, experiment_config_id, experiment_path, writer):
            df_temp_space = get_dataframe(in_space)
            df_temp_space_units = get_dataframe(in_space_units)
            df_temp_space = pd.concat([df_temp_space, df_temp_space_units])
            df_temp_space = df_temp_space.fillna("Not In Space").T
            df_temp_space.columns = ["Space", "Unit"]
            df_temp_space.index.name = "Parameter"
            if experiment_config_id and experiment_path:
                writer.write("\n")
                writer.write(df_temp_space.to_markdown())
                writer.write("\n\n")

        if hasattr(self, "experiment_config_id"):
            experiment_config_id = self.experiment_config_id
            experiment_path = self.experiment_path  # type: ignore
        else:
            experiment_config_id = None
            experiment_path = None

        def write_details_start(writer, string):
            writer.write("\n")
            writer.write(f"/// details | Expand to see the {string}\n")
            writer.write("    type: note\n")
            writer.write("    open: false\n")
            writer.write("\n")

        def write_details_end(writer):
            writer.write("\n\n")
            writer.write("///\n\n")

        # Process the observations and actions for each agen
        #
        file_path = f"{experiment_path}/{experiment_config_id}.md"
        with open(file_path, "w") as writer:
            for agent_key, agent_value in tmp._agent_dict.items():  # noqa: SLF001
                # for debug runs with both agents the same just skip the red policies
                writer.write(f"#### {agent_key[agent_key.find('_') + 1 :].upper()} Agent\n")
                writer.write("\n")

                # reward_list = agent_value.agent_reward_functors
                # if reward_list:
                #     write_details_start(writer, "Rewards")
                #     df_reward = get_dataframe({r.name: print_fields(dict(r.config)) for r in reward_list}).T
                #     writer.write(df_reward.T.to_markdown())
                #     write_details_end(writer)

                write_details_start(writer, "Action Space")
                process_space(
                    agent_value.action_space, "Action Space", agent_value.action_units, experiment_config_id, experiment_path, writer
                )
                write_details_end(writer)

                write_details_start(writer, "Observation Space")
                process_space(
                    agent_value.observation_space,
                    "Observation Space",
                    agent_value.observation_units,
                    experiment_config_id,
                    experiment_path,
                    writer,
                )
                write_details_end(writer)

                write_details_start(writer, "Normalized Action Space")
                process_space(
                    agent_value.normalized_action_space,
                    "Action Space",
                    agent_value.action_units,
                    experiment_config_id,
                    experiment_path,
                    writer,
                )
                write_details_end(writer)

                write_details_start(writer, "Normalized Observation Space")
                process_space(
                    agent_value.normalized_observation_space,
                    "Observation Space",
                    agent_value.observation_units,
                    experiment_config_id,
                    experiment_path,
                    writer,
                )
                write_details_end(writer)

            # Process the evnironment dones
            writer.write("#### Task Termination / Goal Criteria -\n")
            writer.write("\n")

            for index, value in enumerate(tmp._shared_done.process_callbacks):  # noqa: SLF001
                writer.write(f"/// tab | {value.name}\n")  # type: ignore
                if index == 0:
                    writer.write("    new: true\n")
                writer.write("\n")
                writer.write("```yaml\n")
                writer.write(json.dumps(dict(value.config), indent=4, sort_keys=True, cls=QuantityEncoder))  # type: ignore
                writer.write("\n")
                writer.write("```\n")
                writer.write("\n")
                writer.write("///\n\n")

            writer.write("#### Platform Termination / Goal Criteria -\n")
            writer.write("\n")

            for key, dones in tmp._done.items():  # noqa: SLF001
                writer.write(f"##### {key} -\n")
                writer.write("\n")

                for index, value in enumerate(dones):
                    writer.write(f"/// tab | {value.name}\n")  # type: ignore
                    if index == 0:
                        writer.write("    new: true\n")
                    writer.write("\n")
                    writer.write("```yaml\n")
                    writer.write(json.dumps(dict(value.config), indent=4, sort_keys=True, cls=QuantityEncoder))  # type: ignore
                    writer.write("\n")
                    writer.write("```\n")
                    writer.write("\n")
                    writer.write("///\n\n")

    def setup_configs(self, args):
        """
        tbd
        """
        rllib_config = self._select_rllib_config(args.compute_platform)
        if args.compute_platform in ["ray"]:
            self._update_ray_config_for_ray_platform()

        self.config.env_config["agents"], self.config.env_config["agent_platforms"] = self.create_agents(
            args.platform_config,
            args.agent_config,
        )

        self.config.env_config["horizon"] = rllib_config["horizon"]

    def _select_rllib_config(self, platform: str | None) -> dict[str, typing.Any]:
        """Extract the rllib config for the proper computational platform.

        Parameters
        ----------
        platform : typing.Optional[str]
            Specification of the computational platform to use, such as "local", "hpc", etc.  This must be present in the rllib_configs.
            If None, the rllib_configs must only have a single entry.

        Returns
        -------
        typing.Dict[str, typing.Any]
            Rllib configuration for the desired computational platform.

        Raises
        ------
        RuntimeError
            The requested computational platform does not exist or None was used when multiple platforms were defined.
        """
        if platform is not None:
            return self.config.rllib_configs[platform]

        if len(self.config.rllib_configs) == 1:
            return self.config.rllib_configs[next(iter(self.config.rllib_configs))]

        raise RuntimeError(f'Invalid rllib_config for platform "{platform}"')

    def _update_ray_config_for_ray_platform(self) -> None:
        """Update the ray configuration for ray platforms."""
        self.config.ray_config["address"] = "auto"
        self.config.ray_config["log_to_driver"] = False



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import argparse
import logging
import pathlib
import typing
import warnings

import numpy as np

from corl.experiments.base_experiment import BaseExperiment, ExperimentFileParse, ExperimentParse
from corl.libraries.config_file_watcher import LoggingSetup
from corl.parsers.yaml_loader import load_file

warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)


def merge_cfg_and_args(cfg: dict, args: argparse.Namespace) -> dict:
    """
    override items in the cfg file with arguments from command line
    """
    for arg_name in ["compute_platform", "environment", "name", "output"]:
        if arg_val := getattr(args, arg_name):
            cfg[arg_name] = arg_val
    if args.debug:
        cfg["debug"] = True
    return cfg


def build_experiment(args) -> tuple[BaseExperiment, ExperimentFileParse]:
    """
    handles the argparsing for CoRL, this function allows alternate arguments to be
    input to allow unit tests to use the same parsing code

    Args:
        alternate_argv (typing.Optional[typing.Sequence[str]], optional): _description_. Defaults to None.

    Returns:
        _type_: fully parsed arguments
    """
    cfg = load_file(config_filename=args.cfg)
    cfg = merge_cfg_and_args(cfg, args)
    experiment_file_validated = ExperimentFileParse(**cfg)
    config = load_file(config_filename=str(experiment_file_validated.config))
    experiment_parse = ExperimentParse(**config)
    experiment_parse.experiment_class.process_cli_args(experiment_parse.config, experiment_file_validated)
    experiment_class = experiment_parse.experiment_class(**experiment_parse.config)

    return experiment_class, experiment_file_validated


def parse_corl_args(alternate_argv: typing.Sequence[str] | None = None) -> argparse.Namespace:
    """
    handles the argparsing for CoRL, this function allows alternate arguments to be
    input to allow unit tests to use the same parsing code

    Args:
        alternate_argv (typing.Optional[typing.Sequence[str]], optional): _description_. Defaults to None.

    Returns:
        _type_: fully parsed arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--cfg",
        type=pathlib.Path,
        help="Path to config.yml file used to setup the training environment",
    )
    parser.add_argument(
        "--compute-platform",
        type=str,
        default="default",
        help="Compute platform [ace, hpc, local] of experiment. Used to select rllib_config",
    )
    parser.add_argument(
        "--environment",
        type=str,
        help="Compute environment [env1, env2, default] of experiment. Used to select env_config",
    )
    parser.add_argument(
        "--name", action="store", help="Tells your specified experiment to update its name. Experiments may ignore this directive."
    )
    parser.add_argument(
        "--output",
        action="store",
        help="Tells your specified experiment to update its output directory.  Experiments may ignore this directive.",
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Log level. This will be overridden by --log-config if log config exists.",
    )
    parser.add_argument(
        "--log-config", type=pathlib.Path, default=pathlib.Path.cwd() / "logging.yml", help="Optional path to yaml logging configuration."
    )
    # Canonical way to set flags while support vscode `pickString`
    parser.add_argument("--no-debug", dest="debug", action="store_false")
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Tells your specified experiment to switch configurations to debug mode.  Experiments may ignore this directive",
    )
    parser.set_defaults(debug=False)
    return parser.parse_args(args=alternate_argv)


def main() -> None:
    """
    Main method of the module that allows for arguments parsing  for experiment setup.
    """
    args = parse_corl_args()
    LoggingSetup(default_path=str(args.log_config), default_level=logging._nameToLevel[args.log_level])  # noqa: SLF001
    experiment_class, experiment_file_validated = build_experiment(args)
    experiment_class.run_experiment(experiment_file_validated)


if __name__ == "__main__":
    main()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Pong Game Playback
"""

import datetime
import glob
import os
import pickle
import shutil
from pathlib import Path

import imageio
import jsonargparse
import pandas as pd
import pygame
from pygifsicle import optimize
from tqdm import tqdm

from corl.simulators.pong.pong import GameStatus, PongRender


def create_gif(image_directory: str, output_filename: str, fps: int):
    """Creates a gif of all images in the `image_directory` and saves it
    out to `output_filename`.

    Parameters
    ----------
    image_directory : str
        The directory containing the individual image frames.
    output_filename : str
        Output filename.
    fps : int
        Frames per second.
    """

    writer: imageio.plugins.pillowmulti.GIFFormat.Writer

    image_files = glob.glob(f"{image_directory}/*")
    image_files.sort()
    print("Creating a gif...")
    # Reads one frame in at a time while creating the gif
    with imageio.v2.get_writer(output_filename, duration=1000 / fps) as writer:
        for file in tqdm(image_files):
            image = imageio.v2.imread(file)
            writer.append_data(image)

    # Attempts lossless compression
    optimize(output_filename)


def main(pickle_files: list, parse_args: jsonargparse.Namespace):  # noqa: PLR0915
    """Episode playback. Renders every episode in the list of pickle files
    sequentially. Displays the simulation time, accumulated awards and episode
    file as text.

    Parameters
    ----------
    pickle_files : typing.List
        A list of files to iterate through.
    """

    pygame.init()

    clock = pygame.time.Clock()
    fps = 60
    left_score = 0
    right_score = 0
    pong_render = None
    tmp_directory = "/tmp/pong_frames/"  # noqa: S108
    frame = 0

    # Make sure this directory is empty
    os.makedirs(tmp_directory, exist_ok=False)

    for file in tqdm(pickle_files):
        with open(file, "rb") as file:  # noqa: PLW2901
            episode_states = pickle.load(file)  # noqa: S301

        pong = episode_states[0]["pong_game"]
        pong_render = PongRender(pong.screen_width, pong.screen_height)
        file_to_display = str(file.name).replace(parse_args.input, "...") if os.path.isdir(parse_args.input) else parse_args.input
        text_block = []

        episode_length = len(episode_states)
        # Iterates through an episode
        for state in episode_states:
            clock.tick(fps)

            pong = state["pong_game"]
            # Construct extra text information to show
            text_block.append(file_to_display)
            for key, value in state["rewards_accumulator"].items():
                text_block.append(f"{key}:{value}")
            sim_time = state["sim_time"]
            text_block.append(f"sim time: {sim_time}")

            # Renders each frame
            pong_render.draw(pong, left_score, right_score, text_block)
            game_status = state["game_status"]

            if game_status is GameStatus.RIGHT_WIN:
                right_score += 1
            elif game_status is GameStatus.LEFT_WIN:
                left_score += 1

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    break

            # zero pad the frame count
            frame_count = str(frame).zfill(len(str(episode_length) * len(pickle_files)))
            pygame.image.save(pong_render.display_window, f"{tmp_directory}/{frame_count}.jpg")
            text_block = []
            frame += 1

        # Creates the gif for the episode
        if not parse_args.one_gif_all_episodes:
            output_filename = Path(file.name).name.replace(".pkl", ".gif")
            create_gif(image_directory=tmp_directory, output_filename=f"{parse_args.gif_output_directory}/{output_filename}", fps=fps)
            frame = 0
            # Removes the directory and recreates it
            shutil.rmtree(tmp_directory)
            os.makedirs(tmp_directory, exist_ok=False)

        # Updates the final score
        pong_render.draw(pong, left_score, right_score, text_block)

    # Creates one gif for all the episodes
    if parse_args.one_gif_all_episodes:
        filename_start = Path(pickle_files[0].name).name.replace(".pkl", "")
        filename_end = Path(pickle_files[-1].name).name.replace(".pkl", "")
        output_filename = f"{filename_start}_{filename_end}-{datetime.datetime.now().isoformat()}.gif"

        create_gif(image_directory=tmp_directory, output_filename=f"{parse_args.gif_output_directory}/{output_filename}", fps=fps)
        # Remove the temporary directory
        shutil.rmtree(tmp_directory)

    if pong_render:
        pong_render.draw_win("Finished playback.")

    pygame.quit()
    shutil.rmtree(tmp_directory)


if __name__ == "__main__":
    parser = jsonargparse.ArgumentParser()
    parser.add_argument("--input", help="Top level directory to where the episode states are saved or a single episode.")
    parser.add_argument("--later-episodes-first", help="Plays episodes in reverse order, with newer ones first. ", action="store_true")
    parser.add_argument("--one-gif-all-episodes", action="store_true", help="Creates a single GIF for all episodes in the input.")
    parser.add_argument("--gif-output-directory", default="gifs/")
    args = parser.parse_args()
    if os.path.isdir(args.input):
        episode_files = list(Path(args.input).rglob("episode*.pkl"))
        file_df = pd.DataFrame(episode_files, columns=["full_filepath"])

        file_df["filename"] = file_df["full_filepath"].apply(lambda x: x.name)
        file_df = file_df.sort_values(by=["filename"], ascending=not args.later_episodes_first)
        episode_files = file_df["full_filepath"].tolist()
    else:
        episode_files = [Path(args.input)]

    print(f"{len(episode_files)} episodes found.")
    os.makedirs(args.gif_output_directory, exist_ok=True)
    main(pickle_files=episode_files, parse_args=args)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import collections
import copy
import json
import os
from pathlib import Path
from typing import IO, Any
from urllib.parse import unquote, urlparse

import importlib_metadata
import yaml
from ray import tune
from yaml.constructor import ConstructorError
from yaml.nodes import SequenceNode


def get_pkgs_paths() -> set[Path]:
    """returns set of full paths to all pkgs"""
    pkgs_paths = set()
    for distribution in importlib_metadata.distributions():
        if distribution.origin:
            pkgs_paths.add(Path(unquote(urlparse(distribution.origin.url).path)))
        else:
            pkgs_paths.add(Path(str(distribution.locate_file(""))))
    return pkgs_paths


def get_distribution_path(distribution: importlib_metadata.Distribution) -> Path:
    """returns set of full path to top level of a package"""
    dist_name = importlib_metadata.Prepared.normalize(distribution.name)
    if distribution.origin:
        return Path(unquote(urlparse(distribution.origin.url).path)) / Path(dist_name)
    return (Path(str(distribution.locate_file("")))) / Path(dist_name)


class Loader(yaml.SafeLoader):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream: IO) -> None:
        """Initialise Loader."""

        try:
            self.root_path = Path(os.path.split(stream.name)[0])
        except AttributeError:
            self.root_path = Path(os.path.curdir)

        self.cwd_path = Path.cwd()

        self._pkg_paths = get_pkgs_paths()
        self._distribution = None

        for site_path in self._pkg_paths:
            if site_path in self.root_path.parents:
                str(Path(*self.root_path.relative_to(site_path).parts[:1]))
                if site_path != self.cwd_path:
                    self._distribution = importlib_metadata.distribution(str(Path(*self.root_path.relative_to(site_path).parts[:1])))

        super().__init__(stream)

        self._include_mapping: dict = {}
        self.deep_construct = True

    def build_file_path(self, node, absolute: bool = False):
        """
        handles loading the strings behind the various corl !include
        extensions to handle complex loading

        corl performs a prioritized search in the following order to
        locate a file and build the associated path. The first search
        that returns a valid file will be return the path. Subsequent
        searches will not be performed.

        1. Relative paths
        if an include filepath begins with a '.' the loader will use the file location as a basis
        no searches are performed beyond relative locations of current file.
        file path: config/tasks/docking1d/agents/main_agent.yml
        example1: !include ./glues/glue_set1.yml
        would be resolved to config/tasks/docking1d/agents/glues/glue_set1.yml

        2. CWD paths
        corl will look for files pathed from the current working directory
        example: !include config/tasks/docking1d/agents/glue_set1.yml
        would be resolved to load <path to working directory>/config/tasks/docking1d/agents/glue_set1.yml

        3. Module paths
        corl sets a module path for each yaml conf. The module path is set to the cwd
        path, unless the current yml conf resides in a site_pkgs (i.e. is installed).
        In that case the module path will be set the the site package directory of the module.
        example: /home/user/src/corl/config/tasks/docking1d/agents/glue_set1.yml
        module_path: cwd
        example: opt/conda/lib/python3.10/site-packages/<pkg>/config/tasks/docking1d/agents/glue_set1.yml
        module_path: opt/conda/lib/python3.10/site-packages/<pkg>

        when parsing a module input config include path can now be defined relative the module
        for example in the config opt/conda/lib/python3.10/site-packages/<pkg>/config/sample_config1.yml
        example: !include config/sample_config2.yml
        would be resolved to load opt/conda/lib/python3.10/site-packages/<pkg>/config/sample_config2.yml

        4. Site-package paths
        corl will search site package for includes not found in other paths
        example: !include <pkg>/config/sampple_config1.yml
        would be resolved to load opt/conda/lib/python3.10/<pkg>/config/sampple_config1.yml
        """
        loader_string = self.construct_scalar(node)
        assert isinstance(loader_string, str)

        filename, distribution = self._find_file(loader_string)

        if not filename:
            raise RuntimeError(f"{loader_string} not found when yaml parsing")

        return filename.absolute() if absolute else filename.resolve(), distribution

    def _find_file(self, loader_string) -> tuple[Path | None, importlib_metadata.Distribution | None]:
        """ """
        # Check if file is relative to current file
        if loader_string.startswith("."):
            tmp_path = Path(self.root_path, loader_string)
            if tmp_path.is_file():
                return tmp_path, self._distribution
            return None, None

        # Check if file is in current repo
        cwd_path = Path(self.cwd_path, loader_string)
        if cwd_path.is_file():
            return cwd_path, None

        # Check if file is in current module
        if self._distribution:
            module_path = Path(get_distribution_path(self._distribution), loader_string)
            if module_path.is_file():
                return module_path, self._distribution

        # Check if file is in a different site pkg module
        for pkg_path in self._pkg_paths:
            site_path = Path(pkg_path, loader_string)
            if site_path.is_file():
                return site_path, importlib_metadata.distribution(str(Path(*site_path.relative_to(pkg_path).parts[:1])))

        return None, None

    def construct_str(self, node):
        # Implement custom string handling here
        if Path(node.value).suffix:
            new_file_path, _ = self._find_file(node.value)
            if new_file_path:
                return str(new_file_path)
        return node.value

    def construct_document(self, node):
        data = super().construct_document(node)
        self.deep_construct = True
        return data

    def construct_python_tuple(self, node):
        """Adds in the capability to process tuples in yaml files"""
        return tuple(self.construct_sequence(node))

    def construct_sequence(self, node, deep=False):
        """Construct a sequence from a YAML sequence node

        This method extends yaml.constructor.BaseConstructor.construct_sequence by adding support for children with the tag
        `!include-extend`.  Any object with this tag should be constructable to produce a sequence of objects.  Even though
        `!include-extend` is a tag on the child object, the sequence produced by this child is not added as a single element to the sequence
        being produced by this method.  Rather, the output sequence is extended with this list.  Any children with other tags are appended
        into the list in the same manner as yaml.constructor.BaseConstructor.construct_sequence.

        Examples
        --------
        Loader.add_constructor("!include-extend", construct_include)
        with open("primary.yml", "r") as fp:
            config = yaml.load(fp, Loader)

        <file primary.yml>
        root:
          tree1:
            - apple
            - banana
            - cherry
          tree2:
            - type: int
              value: 3
            - type: float
              value: 3.14
            - type: str
              value: pi
          tree3:
            - date
            - elderberry
            - !include-extend secondary.yml
            - mango

        <file secondary.yml>
        - fig
        - grape
        - honeydew
        - jackfruit
        - kiwi
        - lemon

        The output of the code above is:
        config = {
            'root': {
                'tree1': ['apple', 'banana', 'cherry'],
                'tree2': [
                    {'type': 'int', 'value': 3},
                    {'type': 'float', 'value': 3.14},
                    {'type': 'str', 'value': 'pi'}
                ],
                'tree3': ['date', 'elderberry', 'fig', 'grape', 'honeydew', 'jackfruit', 'kiwi', 'lemon', 'mango']
            }
        }
        """
        if not isinstance(node, SequenceNode):
            return super().construct_sequence(node, deep=deep)

        output = []
        for child in node.value:
            this_output = self.construct_object(child, deep=deep)
            if child.tag == "!include-extend":
                if not isinstance(this_output, collections.abc.Sequence):
                    raise ConstructorError(
                        None,
                        None,
                        f"expected a sequence returned by 'include-extend', but found {type(this_output).__name__}",
                        child.start_mark,
                    )
                output.extend(this_output)
            else:
                output.append(this_output)

        return output

    def flatten_mapping(self, node):
        merge = []
        index = 0
        while index < len(node.value):
            key_node, value_node = node.value[index]
            if key_node.tag == "tag:yaml.org,2002:merge":
                del node.value[index]
                if isinstance(value_node, yaml.MappingNode):
                    self.flatten_mapping(value_node)
                    merge.extend(value_node.value)
                elif isinstance(value_node, yaml.SequenceNode):
                    submerge = []
                    for subnode in value_node.value:
                        if not isinstance(subnode, yaml.MappingNode):
                            raise yaml.ConstructorError(
                                "while constructing a mapping",
                                node.start_mark,
                                f"expected a mapping for merging, but found {subnode.id}",
                                subnode.start_mark,
                            )
                        self.flatten_mapping(subnode)
                        submerge.append(subnode.value)
                    submerge.reverse()
                    for value in submerge:
                        merge.extend(value)
                else:
                    # TODO FIGURE OUT HOW TO DUMP AND ACCESS THE BASE NODE!!!!
                    # if value_node.tag == '!include-direct':
                    #     filename = os.path.realpath(os.path.join(self._root, self.construct_scalar(value_node)))
                    #     d = yaml.dump(self._include_mapping[filename])
                    #     # for k, v in .items():
                    #     #     sk = yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=str(k))
                    #     #     if isinstance(v, int):
                    #     #         sv = yaml.ScalarNode(tag='tag:yaml.org,2002:int', value=str(v))
                    #     #     else:
                    #     #         sv = yaml.ScalarNode(tag='tag:yaml.org,2002:seq', value=str(v))
                    #     #     merge.extend([(sk, sv)])
                    # else:
                    raise yaml.ConstructorError(
                        "while constructing a mapping",
                        node.start_mark,
                        f"expected a mapping or list of mappings for merging, but found {value_node.id}",
                        value_node.start_mark,
                    )
            elif key_node.tag == "tag:yaml.org,2002:value":
                key_node.tag = "tag:yaml.org,2002:str"
                index += 1
            else:
                index += 1
        if merge:
            node.value = merge + node.value


def include_file(filename: Path) -> Any:
    extension = filename.suffix

    with open(filename, encoding="utf-8") as fp:
        if extension in (".yaml", ".yml"):
            return yaml.load(fp, Loader)  # noqa: S506
        elif extension in (".json",):  # noqa: RET505
            return json.load(fp)
        else:
            return "".join(fp.readlines())


def construct_include(loader: Loader, node: yaml.Node) -> Any:
    """Include file referenced at node."""
    filename, _ = loader.build_file_path(node)
    return include_file(filename)


def construct_path(loader: Loader, node: yaml.Node) -> Any:
    """Construct file path associated with node"""
    filename, _ = loader.build_file_path(node)
    return filename


def construct_include_direct(loader: Loader, node: yaml.Node) -> Any:
    """Include file referenced at node."""
    filename, file_module = loader.build_file_path(node)
    extension = filename.suffix

    with open(filename, encoding="utf-8") as fp:
        if extension in (".yaml", ".yml"):
            temp = yaml.load(fp, Loader)  # noqa: S506
            loader._include_mapping[filename] = temp  # noqa: SLF001
            return temp
        elif extension in (".json",):  # noqa: RET505
            return json.load(fp)
        else:
            return "".join(fp.readlines())


def construct_tune_function(loader: Loader, node: yaml.Node) -> Any:
    """Include expression referenced at node."""
    if isinstance(node.value, str) and "tune" in node.value:
        return eval(node.value)  # noqa: S307
    return node.value


def construct_include_arr(loader: Loader, node: yaml.Node) -> Any:
    """Identical to above, but accepts an array and appends results as an array."""

    sequence = loader.construct_sequence(node)
    data: list = []
    for item in sequence:
        filename, _ = loader.build_file_path(item, absolute=True)
        data = data + include_file(filename)
    return data


def construct_grid_search(loader: Loader, node: yaml.Node) -> Any:
    """Include file referenced at node."""
    sequence = loader.construct_sequence(node)
    return tune.grid_search(sequence)


def construct_merge_dict(loader: Loader, node: yaml.Node):
    sequence = loader.construct_sequence(node)
    return apply_patches(sequence)


Loader.add_constructor("!include", construct_include)
Loader.add_constructor("!include-direct", construct_include_direct)
Loader.add_constructor("!include-extend", construct_include)
Loader.add_constructor("!function", construct_tune_function)
Loader.add_constructor("tag:yaml.org,2002:python/tuple", Loader.construct_python_tuple)
Loader.add_constructor("tag:yaml.org,2002:str", Loader.construct_str)
Loader.add_constructor("!include_arr", construct_include_arr)
Loader.add_constructor("!merge", construct_merge_dict)
Loader.add_constructor("!tune_grid_search", construct_grid_search)
Loader.add_constructor("!corl_path", construct_path)


def load_file(config_filename: str | Path):
    """
    Utility function to load in a specified yaml file
    """
    with open(config_filename, encoding="utf-8") as fp:
        return yaml.load(fp, Loader)  # noqa: S506


def separate_config(config: dict):
    """
    Utility function to separate the env specific configs from the tune configs
    """
    # we can call ray.init without any arguments so default is no arguments
    ray_config = apply_patches(config.get("ray_config", {}))

    # we must have a tune config or else we cannot call tune.run
    if "tune_config" not in config:
        raise ValueError(f"Could not find a tune_config in {config}")
    tune_config = apply_patches(config["tune_config"])

    # must also get an env_config or else we don't know which environment to run```
    if "env_config" in config:
        env_config = apply_patches(config["env_config"])
    else:
        raise ValueError(f"Could not find a env_config in {config} or rllib_config")

    # must get a rllib config in some way or else we aren't going to run anything useful
    rllib_configs = {}
    if "rllib_configs" in config:
        for key, value in config["rllib_configs"].items():
            rllib_configs[key] = apply_patches(value)
    else:
        raise ValueError(f"Could not find a rllib_config in {config} or 'config' in tune_config")

    for value in rllib_configs.values():
        value["env_config"] = copy.deepcopy(env_config)
        value["env_config"].setdefault("environment", {})["horizon"] = value.get("horizon", 1000)

    # a trainable config is not necessary
    trainable_config = apply_patches(config.get("trainable_config", None))

    return ray_config, rllib_configs, tune_config, trainable_config


def apply_patches(config):
    """updates the base setup with patches

    Arguments:
        config [dict, list] -- The base and patch if list, else dict

    Returns:
        The combined dict
    """

    def merge(source, destination):
        """
        run me with nosetests --with-doctest file.py

        >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
        >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
        >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
        True
        """
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, {})
                merge(value, node)
            else:
                destination[key] = value

        return destination

    if isinstance(config, list):
        config_new = copy.deepcopy(config[0])
        for item in config[1:]:
            if item is not None:
                config_new = merge(item, config_new)
        return config_new

    return config



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module defines the TestCaseStrategy for the TabularParameterProvider

Author: John McCarroll
"""


import pandas as pd
from pydantic import BaseModel, ConfigDict

from corl.evaluation.runners.section_factories.test_cases.pandas import Pandas
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import (
    TestCaseIndex,
    TestCaseStrategy,
    TestCaseStrategyValidator,
)


class TabularStrategyValidator(TestCaseStrategyValidator):
    """
    A validator for the TabularStrategy class
    """

    separator: str = "."

    config: dict = {}


class PandasValidator(BaseModel):
    """
    A validator for the Pandas class
    """

    data: str | pd.DataFrame
    source_form: str
    seed: int = 12345678903141592653589793
    samples: float | None = None
    randomize: bool = True
    model_config = ConfigDict(arbitrary_types_allowed=True)


class TabularStrategy(TestCaseStrategy):
    """
    The concrete implementation of a TestCaseStrategy for the TabularParameterProvider.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.config: TabularStrategyValidator = self.get_validator()(**kwargs)
        dataframe_config: PandasValidator = self.get_pandas_validator(**self.config.config)
        self._test_cases = self.generate_pandas_dataframe(dataframe_config)  # type: ignore
        self.epp_class_path = "corl.episode_parameter_providers.tabular_parameter_provider.TabularParameterProvider"

    @staticmethod
    def get_validator() -> type[TabularStrategyValidator]:
        """
        Method to return validator for TestCaseStrategy
        """
        return TabularStrategyValidator

    @property
    def get_pandas_validator(self) -> type[PandasValidator]:
        """
        Method to return validator for PandasValidator
        """
        return PandasValidator

    def update_rllib_config(self, config: dict):
        """
        Method responsible for mutating rllib config, overwriting the task's default EPPs with TabularParameterProviders.
        """
        # Override the environment EPP

        # retrieve epp config parse for environment's params
        epp_env_config = self.parse_config_for_eval()

        config["env_config"]["episode_parameter_provider"] = {"type": self.epp_class_path, "config": epp_env_config}

        # Override the agents EPP
        for agent_id, value in config["env_config"]["agents"].items():
            # retrieve epp config parsed for agent's params
            epp_agent_config = self.parse_config_for_eval(agent_id=agent_id)

            value.class_config.config["episode_parameter_provider"] = {"type": self.epp_class_path, "config": epp_agent_config}

        return config

    def parse_config_for_eval(self, agent_id: str | None = None) -> dict:
        """
        A method responsible for generating the config dict for each TabularParameterProvider
        """

        # isolate subset of params for scope of single epp instance
        subset_dataframe = None

        if agent_id:
            # parse epp config for specified agent's params
            agent_columns = [x for x in self._test_cases.columns if x.startswith(agent_id)]
            agent_test_cases = self._test_cases.loc[:, agent_columns]
            subset_dataframe = agent_test_cases.rename(columns={x: x.replace(f"{agent_id}.", "") for x in agent_columns})
        else:
            # parse eval config for environment params
            environment_columns = [x for x in self._test_cases.columns if x.startswith("environment")]
            environment_test_cases = self._test_cases.loc[:, environment_columns]
            subset_dataframe = environment_test_cases.rename(columns={x: x.replace("environment.", "") for x in environment_columns})

        return {"separator": self.config.separator, "data": subset_dataframe}

    def get_test_cases(self):
        """
        Method responsible for retrieving test cases planned for evaluation
        """
        return self._test_cases

    def get_num_test_cases(self) -> int:
        """
        Method responsible for reporting the number of test cases planned for evaluation
        """
        return len(self._test_cases.index)

    def generate_pandas_dataframe(self, config: dict) -> pd.DataFrame:  # noqa: PLR6301
        """
        Method responsible for generating the pandas.DataFrame used to maintain test case param values for TabularParameterProvider
        """

        pandas_manager = Pandas(**dict(config))
        pandas_manager.generate()
        return pandas_manager.data_frame

    def get_test_case_index(self, episode_id: int) -> TestCaseIndex:  # noqa: PLR6301
        """For the TabularEpisodeParameterProvider, the episode_id corresponds to the test_case"""
        return TestCaseIndex(episode_id)



"""
---------------------------------------------------------------------------

This is a US Government Work not subject to copyright protection in the US.
The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import streamlit as st
import yaml
from pandas.api.types import is_categorical_dtype, is_numeric_dtype
from ray.rllib.env.env_context import EnvContext
from ray.rllib.policy.sample_batch import SampleBatch
from stqdm import stqdm

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.runners.section_factories.engine.rllib.episode_artifact_logging_callback import EVAL_CHECKPOINTS_SUBDIR
from corl.evaluation.runners.section_factories.engine.rllib.rllib_trainer import ray_context
from corl.visualization.network_explainability.network_explainability import AgentCheckpoint
from corl.visualization.network_explainability.network_explainability_plotter import PPOTorchNetworkExplainabilityPlotter

use_container_width = True


def write_code_block(text: str):
    """Helper function to wrap text in code-style block"""
    st.markdown(
        f"""
    ```
    {text}
    ```
    """
    )


def convert_counts_to_percentage(input_table: pd.DataFrame, count_column: str) -> pd.Series:
    """Helper function to convert counts to percentages"""
    return 100.0 * input_table[count_column] / input_table[count_column].sum()


def st_dataframe_pagination(table_key: str, pages_dict: dict, tables_dict: dict | None = None, dataframe: pd.DataFrame | None = None):
    """Streamlit dataframe widget for pandas dataframe with pagination
    This relies on the session state variable `pages_dict` to be
    defined.

    Parameters
    ----------
    table_key : str
        The key corresponding to the tables dict and / or pages_dict
    pages_dict : dict
        A dictionary that stores which page the table view is on
    tables_dict : Optional[dict], optional
        A dictionary of all the tables, optional, by default None
    dataframe : Optional[pd.DataFrame], optional
        The dataframe to show, by default None
    """
    VIEW_LIMIT = 1000
    if tables_dict:
        _dataframe: pd.DataFrame = tables_dict[table_key]

    elif dataframe is not None:
        _dataframe = dataframe
    else:
        raise ValueError("A dataframe or tables_dict needs to be provided.")

    _dataframe = _dataframe.copy()
    _dataframe = _dataframe.reset_index(drop=True)
    total_rows = len(_dataframe)
    if len(_dataframe) < VIEW_LIMIT:
        st.dataframe(_dataframe, use_container_width=use_container_width)

    else:
        hash_id = pd.util.hash_pandas_object(_dataframe).sum()
        max_pages = np.ceil(len(_dataframe) / VIEW_LIMIT)
        current_page = pages_dict.get(table_key, 0)
        offset = np.clip(VIEW_LIMIT * current_page, a_min=0, a_max=total_rows - VIEW_LIMIT)
        subset = _dataframe.loc[offset : offset + VIEW_LIMIT]
        st.dataframe(subset, use_container_width=use_container_width)

        columns = st.columns(6)
        with columns[-3]:
            if st.button(label="Prev Page", key=f"{table_key}_prev_page_{current_page}_{hash_id}"):
                pages_dict[table_key] = max(current_page - 1, 0)
                st.experimental_rerun()
        with columns[-2]:
            if st.button(label="Next Page", key=f"{table_key}_next_page_{current_page}_{hash_id}"):
                pages_dict[table_key] = min(current_page + 1, max_pages - 1)
                st.experimental_rerun()
        with columns[-1]:
            st.markdown(f"Showing: {int(offset)}:{int(offset+VIEW_LIMIT)} of {total_rows}")
    return pages_dict


def confidence_interval(a, which=95, axis=None):
    """Return a percentile range from an array of values."""
    p = 50 - which / 2, 50 + which / 2
    return np.nanpercentile(a, p, axis)


def init_network_explainability_module(
    agent_checkpoints: list[AgentCheckpoint],
    episode_artifact_files: list[Path],
    episode_artifact_list: list[EpisodeArtifact] | None = None,
    env_config_path: str | Path | None = None,
    env_config: dict | EnvContext | None = None,
    verbose=True,
    policy_network_input_map: dict | None = None,
) -> PPOTorchNetworkExplainabilityPlotter:
    """Initializes and returns the network explianability module

    Parameters
    ----------
    agent_checkpoints : List[AgentCheckpoint]
        A list of agent checkpoints
    episode_artifact_files : Optional[List[Union[str, Path]]], optional
        A list of episode artifact files, by default None
    episdoe_artifact_list : Optional[List[EpisodeArtifact]], optional
        A list of episode artifacts already loaded, by default None
    env_config_path : Optional[Union[str, Path]], optional
        The path to the environment configuration, by default None
    env_config : Optional[Union[Dict, EnvContext]], optional
        The env context, already loaded, by default None
    verbose : bool, optional
        When True, verbose output, by default True

    Returns
    -------
    PPOTorchNetworkExplainabilityPlotter
        The network visaulizer class.
    """
    if env_config is None:
        with open(env_config_path, "rb") as file_obj:  # type: ignore
            env_config = pickle.load(file_obj)  # noqa: S301
    with ray_context(local_mode=True, include_dashboard=False, num_cpus=1):
        explainer_viz = PPOTorchNetworkExplainabilityPlotter(agent_checkpoints=agent_checkpoints, env_config=env_config)
        if verbose:
            st.text("Generating Policy Network Visualizations...")

        if episode_artifact_list:
            for episode_artifact in stqdm(episode_artifact_list):
                explainer_viz.append_feat_importance_gradient_based(episode_artifact, policy_network_input_map)
                explainer_viz.append_vf_residuals(episode_artifact)

        elif episode_artifact_files:
            for file in stqdm(episode_artifact_files):
                with open(file, "rb") as file_obj:
                    episode_artifact = pickle.load(file_obj)  # noqa: S301
                explainer_viz.append_feat_importance_gradient_based(episode_artifact, policy_network_input_map)
                explainer_viz.append_vf_residuals(episode_artifact)
        else:
            raise ValueError("Must provide either `episode_artifact_files` or `episode_artifact_list")

    return explainer_viz


def st_directory_picker(title: str) -> tuple[Path, str | None]:
    """Generates a text input for selecting a directory

    Parameters
    ----------
    title : str
    initial_path : Path, optional
        The initial path, by default Path()

    Returns
    -------
    Path
        The selected directory.
    """
    st.markdown("#### " + title)

    columns = st.columns(2)

    with columns[0]:
        selected_path = Path(st.text_input("Select a directory:"))

    sub_dirs = [f.name for f in selected_path.iterdir() if f.is_dir() and (not f.stem.startswith(".") and not f.stem.startswith("__"))]
    chosen_subdir = None
    with columns[1]:
        if sub_dirs:
            chosen_subdir = st.selectbox("Choose Run", sorted(sub_dirs))
        else:
            st.markdown("#")
            st.markdown(":red[No subdirectories found]")

    return selected_path, chosen_subdir


def read_eval_experiment_file(experiment_dir):
    """Returns a dictionary corresponding to the experiment set up and
    the name of the yaml file read"""
    yaml_file = list(Path(experiment_dir).glob("*.y*ml"))
    assert len(yaml_file) == 1
    with open(yaml_file[0], "rb") as stream:
        try:
            experiment_config = yaml.safe_load(stream)
            return experiment_config, yaml_file[0]
        except yaml.YAMLError as exc:
            print(exc)
            return None, None


def create_selection_grid_single_and_self_policy(num_checkpoints, training_run_dir, main_policy):
    """Helper function to create the selection grid for single policy and self-play.

    Parameters
    ----------
    num_checkpoints : int
        Number of checkpoints to inspect
    training_run_dir : Union[str, Path]
        The training run directory.
    main_policy : str
        The name of the main policy

    Returns
    -------
    List[Dict[str, Any]]
        A list of the selections where each selection is a dictionary containing
        an AgentCheckpoint and epoch selected.
    """
    training_run_dir = Path(training_run_dir)
    # The grid will be x rows by 2 columns
    assert training_run_dir.exists(), f"Directory does not exist: {training_run_dir}"
    num_columns = min(num_checkpoints, 2)
    num_rows = int(np.ceil(num_checkpoints / num_columns))

    show_columns = num_columns
    items_displayed = 0
    return_selections = []
    _unqiue_selections = []
    for row in range(num_rows):
        _columns = st.columns(2)
        # Shows at most 2 columns
        for col in range(show_columns):
            with _columns[col]:
                checkpoint_options = _checkpoint_options(training_run_dir)
                checkpoint_selected = st.selectbox("Pick a Checkpoint", checkpoint_options, key=f"checkpoint_{col}_{row}")
            # Keeps track of how many items have already been displayed
            items_displayed += 1
            show_columns = min(num_checkpoints - items_displayed, num_columns)
            if checkpoint_selected:
                checkpoint_dir = training_run_dir.joinpath(EVAL_CHECKPOINTS_SUBDIR, checkpoint_selected, "policies", main_policy)
                if checkpoint_selected not in _unqiue_selections:
                    _unqiue_selections.append(str(checkpoint_selected))
                    return_selections.append(
                        {
                            "agent_checkpoint": AgentCheckpoint(agent_name=main_policy, checkpoint_dir=checkpoint_dir),
                            "epoch_selected": checkpoint_selected,
                        }
                    )
                else:
                    st.text(f"Ignored Selections due to Duplication: {checkpoint_selected}")
    return return_selections


def create_selection_grid_policy_is_agent(num_checkpoints, training_run_dir):
    """Creates the selection grid for PolicyIsAgent. Allows the user to specify
    a policy for each checkpoint.

    Parameters
    ----------
    num_checkpoints : int
        Number of checkpoints to inspect
    training_run_dir : Union[str, Path]
        The training run directory.
    Returns
    -------
    List[Dict[str, Any]]
        A list of the selections where each selection is a dictionary containing
        an AgentCheckpoint and epoch selected.
    """
    training_run_dir = Path(training_run_dir)
    _unique_selections = []
    # The grid will be x rows by 2 columns
    assert training_run_dir.exists(), f"Directory does not exist: {training_run_dir}"
    num_rows = num_checkpoints
    return_selections = []
    for row in range(num_rows):
        _columns = st.columns(2)
        with _columns[0]:
            checkpoint_options = _checkpoint_options(training_run_dir)
            checkpoint_selected = st.selectbox("Pick a Checkpoint", checkpoint_options, key=f"checkpoint_{row}")
        if checkpoint_selected:
            with _columns[1]:
                policy_options = [
                    path.name for path in training_run_dir.joinpath(EVAL_CHECKPOINTS_SUBDIR, checkpoint_selected, "policies").glob("*")
                ]
                policy_options.sort()
                policy_selected = st.selectbox("Pick a policy", policy_options, key=f"policy_{row}")
                checkpoint_dir = None
                if policy_selected:
                    checkpoint_dir = training_run_dir.joinpath(EVAL_CHECKPOINTS_SUBDIR, checkpoint_selected, "policies", policy_selected)
            _selection_key = f"{policy_selected}-{checkpoint_selected}"
            if _selection_key not in _unique_selections and checkpoint_dir:
                return_selections.append(
                    {
                        "agent_checkpoint": AgentCheckpoint(agent_name=policy_selected, checkpoint_dir=checkpoint_dir),
                        "epoch_selected": checkpoint_selected,
                    }
                )
                _unique_selections.append(_selection_key)
            else:
                st.text(f"Ignored Selection due to Duplication, checkpoint: {checkpoint_selected}, policy: {policy_selected}")
    return return_selections


@st.cache_resource
def generate_grad_based_feat_importance(input_fig):
    """Helper function to generate gradient based feature importance"""
    st.markdown("### Feature Importance: Gradient Based")
    st.plotly_chart(input_fig, theme=None, use_container_width=use_container_width)
    st.markdown(
        "**Figure**: Average Normalized Feature Importance: aggregated across all test cases. \
        The gradient-based feature importance depicts the gradient of the network output w.r.t. each network \
            input. The absolute value of the raw gradient-based feature importance is taken before it is averaged \
            and normalized. Observations represented by a Discrete / MultiDiscrete space are transformed into one-hot encoded vectors \
            before they are ingested by RLLib. These vectors are consolidated (summed) in this visualization.  \
            When the policy-network input contains more than the current observation (e.g. a series of observations), \
            the absolute value of the gradients are additionally summed across the time-axis."
    )
    return input_fig


@st.cache_resource
def generate_vf_residuals_dist(input_fig):
    """Helper function to generate gradient based feature importance"""
    st.markdown("### Value Function Residuals")
    st.plotly_chart(input_fig, theme=None, use_container_width=use_container_width)
    st.markdown("**Figure**: The residual of the value function.")
    return input_fig


def _checkpoint_options(training_run_dir):
    """Helper function to show the checkpoint options."""
    checkpoint_options = [path.name for path in training_run_dir.glob(f"{EVAL_CHECKPOINTS_SUBDIR}/epoch_*")]
    checkpoint_options.sort()
    return checkpoint_options


def _check_required_subdirectories_policy_viz(root_dir):
    """Helper function that checks the required subdirectories for the policy
    visualization. Produces a helpful error when a directory is not found."""
    root_dir = Path(root_dir)
    error_msg = f"Run: {root_dir} is missing {EVAL_CHECKPOINTS_SUBDIR}"
    assert root_dir.joinpath(EVAL_CHECKPOINTS_SUBDIR).exists(), error_msg
    error_msg = f"Run: {root_dir} is missing trajectories"
    assert root_dir.joinpath("trajectories").exists(), error_msg


def _group_checkpoints_by_input_output(input_output_dict):
    """Given an input_output_dict with keys checkpoint, input and output,
    groups by input and output and returns grouped checkpoints"""
    dataframe = pd.DataFrame(input_output_dict)
    dataframe["checkpoint"] = dataframe["checkpoint"].apply(lambda row: [row])
    dataframe = dataframe.groupby(["input", "output"]).agg({"checkpoint": sum})
    return dataframe["checkpoint"].values.tolist()


def checkbox_grid(
    possible_choices: list, key_prefix: str, prev_selected: list | None = None, possible_choices_return_value: list | None = None
):
    """
    Creates a block of checkboxes for user selection
    Parameters
    ----------
    possible_choice_list : List
        A list of possible choices
    key_prefix : str
        The prefix for the checkbox key. When this function is called more than once
        a unique key_prefix must be provided each time to satisfy strealmit.
    prev_selected : List, optional
        All values provided in this list will be set to True. Defaults to None.
    Returns
    -------
    List
        A list of selected choices
    """
    if possible_choices_return_value:
        assert len(possible_choices) == len(
            possible_choices_return_value
        ), "The length of possible choices and possible_choices_return_value must be the same."

    if prev_selected is None:
        prev_selected = []

    selected_choices = []
    num_columns = 3
    columns = st.columns(num_columns)
    items_per_column = int(np.ceil(len(possible_choices) / num_columns))
    # Create check boxes. The checkboxes are displayed in `num_columns` columns
    column_counter = 0
    for idx, attribute in enumerate(possible_choices):
        with columns[column_counter]:
            state = st.checkbox(label=attribute, value=attribute in prev_selected, key=f"{key_prefix}_{attribute}")
        if state:
            if possible_choices_return_value:
                selected_choices.append(possible_choices_return_value[idx])
            else:
                selected_choices.append(attribute)
        if (idx + 1) % items_per_column == 0:
            column_counter += 1
    return selected_choices


def center_plotly_figure(figure):
    """Helper function to center the plotly figure."""
    columns = st.columns(3)
    for idx, column in enumerate(columns):
        with column:
            if idx == 1:
                st.plotly_chart(figure, use_container_width=False, theme=None)
            else:
                pass


def text_section_policy_network_input(agent_checkpoints: list[AgentCheckpoint]) -> dict[str, str]:
    """Generates an input section for the policy network input names as free-form text.
    The default value is `SampleBatch.OBS`"""
    st.markdown(
        f"""**Specify the poilcy network inputs for each trained policy.**
    This is typically {SampleBatch.OBS}, for the default model used in `PPOTrainer`."""
    )
    policy_input_map: dict = {}
    n_columns = min(len(agent_checkpoints), 5)
    column_idx = 0
    columns = st.columns(n_columns)

    for agent_checkpoint in agent_checkpoints:
        agent_id = agent_checkpoint.agent_name
        with columns[column_idx]:
            policy_input_map[agent_id] = st.text_input(label=agent_id, key=agent_id, value=SampleBatch.OBS)
        column_idx += 1
        # Once n_columns columns are reached, create a new row
        if column_idx % n_columns == 0:
            column_idx = 0
            columns = st.columns(n_columns)

    st.write(policy_input_map)

    return policy_input_map


def ui_dataframe_filter(dataframe: pd.DataFrame, markdown_title: str | None = None) -> pd.DataFrame:
    """
    Adds a UI on top of a dataframe to let viewers filter columns
    """
    if markdown_title:
        st.markdown(markdown_title)
    dataframe = dataframe.copy()
    with st.container():
        to_filter_columns = st.multiselect("Filter dataframe on", dataframe.columns)
        for column in to_filter_columns:
            _, right_column = st.columns((1, 20))

            # For numeric values, show a slider
            if is_numeric_dtype(dataframe[column]):
                _min = float(dataframe[column].min())
                _max = float(dataframe[column].max())
                step = (_max - _min) / 100
                user_num_input = right_column.slider(
                    f"Values for {column}",
                    min_value=_min,
                    max_value=_max,
                    value=(_min, _max),
                    step=step,
                )
                dataframe = dataframe[dataframe[column].between(*user_num_input)]
            # Treat columns with < 10 unique values as categorical
            elif is_categorical_dtype(dataframe[column]) or dataframe[column].nunique() < 10:
                user_cat_input = right_column.multiselect(
                    f"Values for {column}",
                    dataframe[column].unique(),
                    default=list(dataframe[column].unique()),
                )
                dataframe = dataframe[dataframe[column].isin(user_cat_input)]
            # Otherwise, cast as string and allow users to allow user to use substring or regex
            else:
                user_text_input = right_column.text_input(f"Substring or regex in {column}")
                if user_text_input:
                    dataframe = dataframe[dataframe[column].astype(str).str.contains(user_text_input)]
        st.dataframe(dataframe, use_container_width=True)
    return dataframe



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------
"""
import abc
import random
from functools import partial

from pydantic import BaseModel
from ray import tune


class ParametersPPO:
    """Utility functions for processing hparam searches in the  framework for PPO algorithm
    https://github.com/ray-project/ray/blob/00922817b66ee14ba215972a98f416f3d6fef1ba/rllib/agents/ppo/ppo.py
    https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
    https://blog.openai.com/openai-five/
    https://docs.ray.io/en/master/tune/api_docs/trainable.html#advanced-reusing-actors
    """

    LAMBDA_MIN = 0.9
    LAMBDA_MAX = 1.0
    LAMBDA_DIST = tune.uniform(LAMBDA_MIN, LAMBDA_MAX)

    @staticmethod
    def LAMBDA_RANGE(spec):
        """Sets the default search space for HPARAM
        GAE Parameter Lambda Range: 0.9 to 1
        GAE Parameter Lambda also known as: GAE Parameter (lambda) (PPO Paper), lambda (RLlib),
        lambda (ppo2 baselines), lambda (ppo baselines), lambda (Unity ML), gae_lambda (TensorForce)
        """
        return ParametersPPO.LAMBDA_DIST

    VF_LOSS_COEFF_MIN = 0.5
    VF_LOSS_COEFF_MAX = 1.0
    VF_LOSS_COEFF_DIST = tune.uniform(VF_LOSS_COEFF_MIN, VF_LOSS_COEFF_MAX)

    @staticmethod
    def VF_LOSS_COEFF_RANGE(spec):
        """Sets the default search space for HPARAM
        Value Function Coefficient Range: 0.5, 1
        Value Function Coefficient also known as: VF coeff. (PPO Paper), vf_loss_coef (RLlib),
        vf_coef (ppo2 baselines), (ppo baselines: unclear), (Unity ML: unclear), (TensorForce: unclear)
        """
        return ParametersPPO.VF_LOSS_COEFF_DIST

    ENTROPY_COEFF_MIN = 0.00
    ENTROPY_COEFF_MAX = 0.01
    ENTROPY_COEFF_DIST = tune.uniform(ENTROPY_COEFF_MIN, ENTROPY_COEFF_MAX)

    @staticmethod
    def ENTROPY_COEFF_RANGE(spec):
        """Sets the default search space for HPARAM
        Entropy Coefficient Range: 0 to 0.01
        Entropy Coefficient also known as: Entropy coeff. (PPO Paper), entropy_coeff (RLlib),
        ent_coeff (ppo2 baselines), entcoeff (ppo baselines), beta (Unity ML), entropy_regularization (TensorForce)
        """
        return ParametersPPO.ENTROPY_COEFF_DIST

    CLIP_PARAM_MIN = 0.1
    CLIP_PARAM_MAX = 0.3
    CLIP_PARAM_DIST = tune.choice([0.1, 0.2, 0.3])

    @staticmethod
    def CLIP_PARAM_RANGE(spec):
        """Sets the default search space for HPARAM
        Clipping Range: 0.1, 0.2, 0.3
        Clipping also known as: Clipping parameter epsilon (PPO Paper), clip_param (RLlib),
        cliprange (ppo2 baselines), clip_param (ppo baselines), epsilon (Unity ML),
        likelihood_ratio_clipping (TensorForce)
        """
        return ParametersPPO.CLIP_PARAM_DIST

    KL_TARGET_MIN = 0.003
    KL_TARGET_MAX = 0.03
    KL_TARGET_DIST = tune.uniform(KL_TARGET_MIN, KL_TARGET_MAX)

    @staticmethod
    def KL_TARGET_RANGE(spec):
        """Sets the default search space for HPARAM
        https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
        The KL penalty implementation (third line in the above picture) is available in RLlib's PPO
        implementation. The parameters kl_coeff (initial coefficient for KL divergence) and kl_target
        can be used for the KL implementation.
        KL Target Range: 0.003 to 0.03
        KL Initialization Range: 0.3 to 1  --- KL_COEFF IN RLLIB
        """
        return ParametersPPO.KL_COEFF_DIST

    KL_COEFF_MIN = 0.2  # RLLIB Default
    KL_COEFF_MAX = 1.0  # https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
    KL_COEFF_DIST = tune.uniform(KL_COEFF_MIN, KL_COEFF_MAX)

    @staticmethod
    def KL_COEFF_RANGE(spec):
        """Sets the default search space for HPARAM
        The KL penalty implementation (third line in the above picture) is available in RLlib's PPO
        implementation. The parameters kl_coeff (initial coefficient for KL divergence) and kl_target
        can be used for the KL implementation.
        KL Target Range: 0.003 to 0.03
        KL Initialization Range: 0.3 to 1  --- KL_COEFF IN RLLIB
        """
        return ParametersPPO.KL_COEFF_DIST

    GAMMA_MIN = 0.8000
    GAMMA_MAX = 0.9997
    GAMMA_DIST = tune.uniform(GAMMA_MIN, GAMMA_MAX)

    @staticmethod
    def GAMMA_RANGE(spec):
        """Sets the default search space for HPARAM
        Discount Factor Gamma Range: 0.99 (most common), 0.8 to 0.9997
        Discount Factor Gamma also known as: Discount (gamma) (PPO Paper), gamma (RLlib), gamma (ppo2 baselines),
        gamma (ppo baselines), gamma (Unity ML), discount (TensorForce)
        """
        return ParametersPPO.GAMMA_DIST

    LR_MIN = 5e-6
    LR_MAX = 0.003
    LR_DIST = tune.uniform(LR_MIN, LR_MAX)

    @staticmethod
    def LR_RANGE(spec):
        """Sets the default search space for HPARAM
        Learning Rate Range: 0.003 to 5e-6
        Learning Rate also known as: Adam stepsize (PPO Paper), sgd_stepsize (RLlib), lr (ppo2 baselines),
        (ppo baselines: unclear), learning_rate (Unity ML), learning_rate (TensorForce)
        """
        return ParametersPPO.LR_DIST

    NSGD_MIN = 3
    NSGD_MAX = 30
    NSGD_DIST = tune.choice(list(range(NSGD_MIN, NSGD_MAX + 1)))

    @staticmethod
    def NSGD_RANGE(spec):
        """Sets the default search space for HPARAM
        https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
        Epoch Range: 3 to 30
        Epoch also known as: Num. epochs (PPO paper), num_sgd_iter (RLlib), epochs (ppo2 baselines),
        optim_epochs (ppo baselines), num_epoch (Unity ML), (TensorForce: unclear)
        """
        return ParametersPPO.NSGD_DIST

    SGD_MINIBATCH_SIZE_MIN = 128
    SGD_MINIBATCH_SIZE_MAX = 4096
    SGD_MINIBATCH_SIZE_DIST = tune.choice([128, 256, 512, 1024, 2048, 4096])

    @staticmethod
    def SGD_MINIBATCH_SIZE_RANGE(spec):
        """Sets the default search space for HPARAM"""
        return ParametersPPO.SGD_MINIBATCH_SIZE_DIST

    TRAIN_BATCH_SIZE_MIN = 4096
    TRAIN_BATCH_SIZE_MAX = 160000
    TRAIN_BATCH_SIZE_INC = 256

    # [4096, 4352, 4608, 4864, 5120, 5376, 5632, 5888, 6144, 6400, 6656, 6912, 7168, 7424, 7680, 7936, 8192, 8448,
    #  8704, 8960, 9216, 9472, 9728, 9984, 10240, 10496, 10752, 11008, 11264, 11520, 11776, 12032, 12288, 12544,
    #  12800, 13056, 13312, 13568, 13824, 14080, 14336, 14592, 14848, 15104, 15360, 15616, 15872, 16128, 16384,
    #  16640, 16896, 17152, 17408, 17664, 17920, 18176, 18432, 18688, 18944, 19200, 19456, 19712, 19968, 20224,
    #  20480, 20736, 20992, 21248, 21504, 21760, 22016, 22272, 22528, 22784, 23040, 23296, 23552, 23808, 24064,
    #  24320, 24576, 24832, 25088, 25344, 25600, 25856, 26112, 26368, 26624, 26880, 27136, 27392, 27648, 27904,
    #  28160, 28416, 28672, 28928, 29184, 29440, 29696, 29952, 30208, 30464, 30720, 30976, 31232, 31488, 31744,
    #  32000, 32256, 32512, 32768, 33024, 33280, 33536, 33792, 34048, 34304, 34560, 34816, 35072, 35328, 35584,
    #  35840, 36096, 36352, 36608, 36864, 37120, 37376, 37632, 37888, 38144, 38400, 38656, 38912, 39168, 39424,
    #  39680, 39936, 40192, 40448, 40704, 40960, 41216, 41472, 41728, 41984, 42240, 42496, 42752, 43008, 43264,
    #  43520, 43776, 44032, 44288, 44544, 44800, 45056, 45312, 45568, 45824, 46080, 46336, 46592, 46848, 47104,
    #  47360, 47616, 47872, 48128, 48384, 48640, 48896, 49152, 49408, 49664, 49920, 50176, 50432, 50688, 50944,
    #  51200, 51456, 51712, 51968, 52224, 52480, 52736, 52992, 53248, 53504, 53760, 54016, 54272, 54528, 54784,
    #  55040, 55296, 55552, 55808, 56064, 56320, 56576, 56832, 57088, 57344, 57600, 57856, 58112, 58368, 58624,
    #  58880, 59136, 59392, 59648, 59904, 60160, 60416, 60672, 60928, 61184, 61440, 61696, 61952, 62208, 62464,
    #  62720, 62976, 63232, 63488, 63744, 64000, 64256, 64512, 64768, 65024, 65280, 65536, 65792, 66048, 66304,
    #  66560, 66816, 67072, 67328, 67584, 67840, 68096, 68352, 68608, 68864, 69120, 69376, 69632, 69888, 70144,
    #  70400, 70656, 70912, 71168, 71424, 71680, 71936, 72192, 72448, 72704, 72960, 73216, 73472, 73728, 73984,
    #  74240, 74496, 74752, 75008, 75264, 75520, 75776, 76032, 76288, 76544, 76800, 77056, 77312, 77568, 77824,
    #  78080, 78336, 78592, 78848, 79104, 79360, 79616, 79872, 80128, 80384, 80640, 80896, 81152, 81408, 81664,
    #  81920, 82176, 82432, 82688, 82944, 83200, 83456, 83712, 83968, 84224, 84480, 84736, 84992, 85248, 85504,
    #  85760, 86016, 86272, 86528, 86784, 87040, 87296, 87552, 87808, 88064, 88320, 88576, 88832, 89088, 89344,
    #  89600, 89856, 90112, 90368, 90624, 90880, 91136, 91392, 91648, 91904, 92160, 92416, 92672, 92928, 93184,
    #  93440, 93696, 93952, 94208, 94464, 94720, 94976, 95232, 95488, 95744, 96000, 96256, 96512, 96768, 97024,
    #  97280, 97536, 97792, 98048, 98304, 98560, 98816, 99072, 99328, 99584, 99840, 100096, 100352, 100608, 100864,
    #  101120, 101376, 101632, 101888, 102144, 102400, 102656, 102912, 103168, 103424, 103680, 103936, 104192,
    #  104448, 104704, 104960, 105216, 105472, 105728, 105984, 106240, 106496, 106752, 107008, 107264, 107520,
    #  107776, 108032, 108288, 108544, 108800, 109056, 109312, 109568, 109824, 110080, 110336, 110592, 110848,
    #  111104, 111360, 111616, 111872, 112128, 112384, 112640, 112896, 113152, 113408, 113664, 113920, 114176,
    #  114432, 114688, 114944, 115200, 115456, 115712, 115968, 116224, 116480, 116736, 116992, 117248, 117504,
    #  117760, 118016, 118272, 118528, 118784, 119040, 119296, 119552, 119808, 120064, 120320, 120576, 120832,
    #  121088, 121344, 121600, 121856, 122112, 122368, 122624, 122880, 123136, 123392, 123648, 123904, 124160,
    #  124416, 124672, 124928, 125184, 125440, 125696, 125952, 126208, 126464, 126720, 126976, 127232, 127488,
    #  127744, 128000, 128256, 128512, 128768, 129024, 129280, 129536, 129792, 130048, 130304, 130560, 130816,
    #  131072, 131328, 131584, 131840, 132096, 132352, 132608, 132864, 133120, 133376, 133632, 133888, 134144,
    #  134400, 134656, 134912, 135168, 135424, 135680, 135936, 136192, 136448, 136704, 136960, 137216, 137472,
    #  137728, 137984, 138240, 138496, 138752, 139008, 139264, 139520, 139776, 140032, 140288, 140544, 140800,
    #  141056, 141312, 141568, 141824, 142080, 142336, 142592, 142848, 143104, 143360, 143616, 143872, 144128,
    #  144384, 144640, 144896, 145152, 145408, 145664, 145920, 146176, 146432, 146688, 146944, 147200, 147456,
    #  147712, 147968, 148224, 148480, 148736, 148992, 149248, 149504, 149760, 150016, 150272, 150528, 150784,
    #  151040, 151296, 151552, 151808, 152064, 152320, 152576, 152832, 153088, 153344, 153600, 153856, 154112,
    #  154368, 154624, 154880, 155136, 155392, 155648, 155904, 156160, 156416, 156672, 156928, 157184, 157440,
    #  157696, 157952, 158208, 158464, 158720, 158976, 159232, 159488, 159744, 160000]
    TRAIN_BATCH_SIZE_DIST = tune.choice(list(range(2**12, 160000 + TRAIN_BATCH_SIZE_INC, TRAIN_BATCH_SIZE_INC)))

    @staticmethod
    def TRAIN_BATCH_SIZE_RANGE(spec):
        """Sets the default search space for HPARAM"""
        return ParametersPPO.TRAIN_BATCH_SIZE_DIST

    @staticmethod
    def ppo_hyperparameters() -> dict:
        """PPO hyper parameters for hparam search
        https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe

        Returns:
            dict -- model configuration
        """
        ppo_hparams = {}
        ppo_hparams["lambda"] = tune.sample_from(ParametersPPO.LAMBDA_RANGE)
        ppo_hparams["vf_loss_coeff"] = tune.sample_from(ParametersPPO.VF_LOSS_COEFF_RANGE)
        ppo_hparams["entropy_coeff"] = tune.sample_from(ParametersPPO.ENTROPY_COEFF_RANGE)
        ppo_hparams["clip_param"] = tune.sample_from(ParametersPPO.CLIP_PARAM_RANGE)
        ppo_hparams["gamma"] = tune.sample_from(ParametersPPO.GAMMA_RANGE)
        ppo_hparams["lr"] = tune.sample_from(ParametersPPO.LR_RANGE)
        ppo_hparams["num_sgd_iter"] = tune.sample_from(ParametersPPO.NSGD_RANGE)
        ppo_hparams["sgd_minibatch_size"] = tune.sample_from(ParametersPPO.SGD_MINIBATCH_SIZE_RANGE)
        ppo_hparams["train_batch_size"] = tune.sample_from(ParametersPPO.TRAIN_BATCH_SIZE_RANGE)
        ppo_hparams["kl_coeff"] = tune.sample_from(ParametersPPO.KL_COEFF_RANGE)
        ppo_hparams["kl_target"] = tune.sample_from(ParametersPPO.KL_TARGET_RANGE)
        return ppo_hparams

    @staticmethod
    def sample_ppo_hyperparameters() -> dict:
        """PPO hyper parameters for hparam search
        https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe

        Returns:
            dict -- model configuration
        """
        ppo_hparams = ParametersPPO.ppo_hyperparameters()
        for k, v in ppo_hparams.items():
            ppo_hparams[k] = v.sample()
        return ppo_hparams

    @staticmethod
    def pbt_ppo_explore(config: dict) -> dict:
        """The following function links to the companion function above. Sets the clipping needed by PBT
        https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
        Arguments:
            config {dict} -- input config

        Returns:
            dict -- clipped config
        """

        def clip_parameter(config, parameter, parameter_max, parameter_min):
            if config[parameter] > parameter_max:
                config[parameter] = parameter_max
            elif config[parameter] < parameter_min:
                config[parameter] = parameter_min

        clip_parameter(config, "lambda", ParametersPPO.LAMBDA_MAX, ParametersPPO.LAMBDA_MIN)
        clip_parameter(config, "vf_loss_coeff", ParametersPPO.VF_LOSS_COEFF_MAX, ParametersPPO.VF_LOSS_COEFF_MIN)
        clip_parameter(config, "entropy_coeff", ParametersPPO.ENTROPY_COEFF_MAX, ParametersPPO.ENTROPY_COEFF_MIN)
        clip_parameter(config, "gamma", ParametersPPO.GAMMA_MAX, ParametersPPO.GAMMA_MIN)
        clip_parameter(config, "clip_param", ParametersPPO.CLIP_PARAM_MAX, ParametersPPO.CLIP_PARAM_MIN)
        clip_parameter(config, "lr", ParametersPPO.LR_MIN, ParametersPPO.LR_MAX)
        clip_parameter(config, "kl_coeff", ParametersPPO.KL_COEFF_MIN, ParametersPPO.KL_COEFF_MAX)
        clip_parameter(config, "kl_target", ParametersPPO.KL_TARGET_MIN, ParametersPPO.KL_TARGET_MAX)

        sgd_minibatch_size_str = "sgd_minibatch_size"
        train_batch_size_str = "train_batch_size"
        num_sgd_iter_str = "num_sgd_iter"
        clip_parameter(config, num_sgd_iter_str, ParametersPPO.NSGD_MAX, ParametersPPO.NSGD_MIN)
        config[num_sgd_iter_str] = int(config[num_sgd_iter_str])
        clip_parameter(config, sgd_minibatch_size_str, ParametersPPO.SGD_MINIBATCH_SIZE_MAX, ParametersPPO.SGD_MINIBATCH_SIZE_MIN)
        config[sgd_minibatch_size_str] = int(config[sgd_minibatch_size_str])
        clip_parameter(config, train_batch_size_str, ParametersPPO.TRAIN_BATCH_SIZE_MAX, ParametersPPO.TRAIN_BATCH_SIZE_MIN)
        if config[train_batch_size_str] < config[sgd_minibatch_size_str] * 2:
            config[train_batch_size_str] = config[sgd_minibatch_size_str] * 2
        config[train_batch_size_str] = int(config[train_batch_size_str])
        return config


class ParametersModel:
    """Holds the model parameters"""

    FC_LAYER_CHOICES = [32, 64, 128, 256, 512]
    FC_LAYER_COUNT = [2, 3, 4, 5, 6]

    @staticmethod
    def __get_layers(layer_count, FC_FILTER_LOWER_VALUES_THRESHOLD, layer_choices, MIN_LAYER_INDEX) -> list:
        model_layers: list = []
        for _ in range(layer_count):
            temp_layers = layer_choices[MIN_LAYER_INDEX:] if len(model_layers) < FC_FILTER_LOWER_VALUES_THRESHOLD else layer_choices
            if model_layers:
                model_layers.append(random.choice([x for x in temp_layers if x <= model_layers[-1]]))  # noqa: S311
            else:
                model_layers.append(random.choice(temp_layers))  # noqa: S311
        return model_layers

    @staticmethod
    def select_lstm_model() -> dict:
        """[summary]

        Returns
        -------
        dict
            [description]
        """
        model_config = ParametersModel.select_fully_connected_model()
        model_config["use_lstm"] = True
        model_config["max_seq_len"] = random.choice([2, 3, 5, 10])  # noqa: S311
        model_config["lstm_cell_size"] = random.choice([64, 128, 256, 512, 1024, 2048])  # noqa: S311
        model_config["vf_share_layers"] = True
        # model_config["vf_share_layers"] = random.choice([True, False])
        # model_config["lstm_use_prev_action"] = random.choice([True, False])
        # model_config["lstm_use_prev_reward"] = random.choice([True, False])
        # model_config["_time_major"] = random.choice([True, False])

        return model_config

    @staticmethod
    def select_fully_connected_model() -> dict:
        """[summary]

        Returns
        -------
        dict
            [description]
        """
        layer_count = random.choice(ParametersModel.FC_LAYER_COUNT)  # noqa: S311
        layer_choices = ParametersModel.FC_LAYER_CHOICES
        FC_FILTER_LOWER_VALUES_THRESHOLD = 2
        MIN_LAYER_INDEX = 2
        model_layers = ParametersModel.__get_layers(layer_count, FC_FILTER_LOWER_VALUES_THRESHOLD, layer_choices, MIN_LAYER_INDEX)

        model_config: dict = {}
        model_config["fcnet_hiddens"] = model_layers
        model_config["fcnet_activation"] = random.choice(["relu", "tanh"])  # noqa: S311

        return model_config

    @staticmethod
    def select_framestacking_model() -> dict:
        """[summary]

        Returns:
            dict -- [description]
        """
        model_config = ParametersModel.select_fully_connected_model()
        model_config["custom_model"] = "TorchFrameStack"
        model_config["custom_model_config"] = {}
        model_config["custom_model_config"]["num_frames"] = random.choice(list(range(1, 11)))  # noqa: S311
        model_config["custom_model_config"]["include_actions"] = random.choice([True, False])  # noqa: S311
        model_config["custom_model_config"]["include_rewards"] = random.choice([True, False])  # noqa: S311
        layer_count = random.choice(ParametersModel.FC_LAYER_COUNT)  # noqa: S311
        layer_choices = ParametersModel.FC_LAYER_CHOICES
        FC_FILTER_LOWER_VALUES_THRESHOLD = 2
        MIN_LAYER_INDEX = 2
        model_config["custom_model_config"]["post_fcnet_hiddens"] = ParametersModel.__get_layers(
            layer_count, FC_FILTER_LOWER_VALUES_THRESHOLD, layer_choices, MIN_LAYER_INDEX
        )

        return model_config

    @staticmethod
    def select_gtrxl_model() -> dict:
        """[summary]

        Returns
        -------
        dict
            [description]
        """
        model_config = ParametersModel.select_fully_connected_model()
        model_config["use_attention"] = False
        model_config["attention_num_transformer_units"] = random.choice(list(range(1, 6)))  # noqa: S311
        model_config["attention_dim"] = random.choice([64, 128, 256, 512, 1024, 2048])  # noqa: S311
        model_config["attention_num_heads"] = random.choice(list(range(1, 6)))  # noqa: S311
        model_config["attention_head_dim"] = random.choice([64, 128, 256, 512, 1024, 2048])  # noqa: S311
        model_config["attention_memory_inference"] = 50
        model_config["attention_memory_training"] = 50
        model_config["attention_position_wise_mlp_dim"] = random.choice([64, 128, 256, 512, 1024, 2048])  # noqa: S311
        # model_config["attention_init_gru_gate_bias"] = 2.0
        model_config["attention_use_n_prev_actions"] = random.choice(list(range(1, 11)))  # noqa: S311
        model_config["attention_use_n_prev_rewards"] = random.choice(list(range(1, 11)))  # noqa: S311
        return model_config

    @staticmethod
    def select_model(model_choices) -> dict:
        """The following function provides the start to exploring model configurations."""
        model_config_func = random.choice(model_choices)  # noqa: S311

        return model_config_func()


class BaseHparamSearchValidator(BaseModel):
    """
    Base Validator to subclass for Experiments subclassing BaseExperiment
    """


class BaseHparamSearch(abc.ABC):
    """
    Experiment provides an anstract class to run specific types of experiments
    this allows users to do specific setup steps or to run some sort of custom training
    loop
    """

    def __init__(self, **kwargs) -> None:
        self.config: BaseHparamSearchValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseHparamSearchValidator]:
        """Gets the validator

        Returns
        -------
        typing.Type[BaseHparamSearchValidator]
            The validator class
        """
        return BaseHparamSearchValidator

    @abc.abstractmethod
    def add_algorithm_hparams(self, rllib_config: dict, tune_config: dict) -> None:
        """Updates the configurations based on search alg

        Parameters
        ----------
        rllib_config : dict
            The rllib config
        tune_config : dict
            The tune config
        """


class HparamSearchValidator_Shared(BaseHparamSearchValidator):
    """
    Base Validator to subclass for search subclassing
    """

    # "The training result objective value attribute. Stopping procedures will use this attribute."
    metric: str = "episode_reward_mean"
    # One of {min, max}. Determines whether objective is minimizing or maximizing the metric attribute.
    mode: str = "max"
    # A training result attr to use for comparing time. Note that you can pass in something
    # non-temporal such as training_iteration as a measure of progress, the only requirement is
    # that the attribute should increase monotonically.
    time_attr: str = "timesteps_total"
    # The number of samples to collect during HPARAM search (trials)
    samples: int = 4


class HparamSearchValidator_PBT(HparamSearchValidator_Shared):
    """
    Base Validator to subclass for search subclassing
    """

    # The probability of resampling from the original distribution when applying hyperparam_mutations.
    # If not resampled, the value will be perturbed by a factor of 1.2 or 0.8 if continuous, or changed
    # to an adjacent value if discrete. Note that resample_probability by default is 0.25, thus
    # hyperparameter with a distribution may go out of the specific range.
    resample_probability: float = 0.25
    # (float) - Models will be considered for perturbation at this interval of time_attr. Note that
    # perturbation incurs checkpoint overhead, so you shouldn't set this to be too frequent.
    perturbation_interval: float = 4
    # (float) - Models will not be considered for perturbation before this interval of time_attr has
    # passed. This guarantees that models are trained for at least a certain amount of time or timesteps
    # before being perturbed.
    burn_in_period: float = 10


class HparamSearchPPO_PBT(BaseHparamSearch):
    """PPO PBT Search Space
    https://medium.com/aureliantactics/ppo-hyperparameters-and-ranges-6fc2d29bccbe
    """

    @staticmethod
    def get_validator() -> type[HparamSearchValidator_PBT]:
        """gets the configuration for AHBS

        Returns
        -------
        typing.Type[HparamSearchValidator_AHBS]
            validator
        """
        return HparamSearchValidator_PBT

    def add_algorithm_hparams(self, rllib_config, tune_config) -> None:
        """Adds population based training to the configuration (TBD items to be added - default never add)

        Parameters
        ----------
        rllib_config : dict
            The experiment configuration
        tune_config : dict
            The tune configuration
        """

        # Postprocess the perturbed config to ensure it's still valid
        pbt = tune.schedulers.PopulationBasedTraining(
            time_attr=self.config.time_attr,  # type: ignore
            metric=self.config.metric,  # type: ignore
            mode=self.config.mode,  # type: ignore
            perturbation_interval=self.config.perturbation_interval,  # type: ignore
            resample_probability=self.config.resample_probability,  # type: ignore
            burn_in_period=self.config.burn_in_period,  # type: ignore
            log_config=True,
            # Specifies the mutations of these hyper params
            hyperparam_mutations={
                "lambda": ParametersPPO.LAMBDA_DIST,
                "clip_param": ParametersPPO.CLIP_PARAM_DIST,
                "lr": ParametersPPO.LR_DIST,
                "num_sgd_iter": ParametersPPO.NSGD_DIST,
                "sgd_minibatch_size": ParametersPPO.SGD_MINIBATCH_SIZE_DIST,
                "train_batch_size": ParametersPPO.TRAIN_BATCH_SIZE_DIST,
                "vf_loss_coeff": ParametersPPO.VF_LOSS_COEFF_DIST,
                "entropy_coeff": ParametersPPO.ENTROPY_COEFF_DIST,
                "gamma": ParametersPPO.GAMMA_DIST,
                "kl_coeff": ParametersPPO.KL_COEFF_DIST,
                "kl_target": ParametersPPO.KL_TARGET_DIST,
            },
            custom_explore_fn=ParametersPPO.pbt_ppo_explore,
        )
        # These params start off randomly drawn from a set.
        tune_config["scheduler"] = pbt
        tune_config["num_samples"] = self.config.samples  # type: ignore
        rllib_config.update(ParametersPPO.ppo_hyperparameters())


class HparamSearchValidator_AHBS(HparamSearchValidator_Shared):
    """
    Base Validator to subclass for search subclassing
    """

    # max time units per trial. Trials will be stopped after max_t time units (determined
    # by time_attr) have passed.
    max_t: float = 1e7
    # Brackets
    brackets: float = 1
    # Only stop trials at least this old in time. The units are the same as the attribute
    # named by time-attr.
    grace_period: float = 5e6

    include_lstm_search: bool = False
    inclue_fully_connected_search: bool = True
    include_frame_stacking_search: bool = False
    include_gtrxl_search: bool = False


class HparamSearchPPO_AHBS(BaseHparamSearch):
    """Asynchronous Hyper Band Example
    https://docs.ray.io/en/master/tune/examples/includes/async_hyperband_example.html
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(kwargs=kwargs)

        self._model_choices = []

        if self.config.include_lstm_search:  # type: ignore
            self._model_choices.append(ParametersModel.select_lstm_model)

        if self.config.inclue_fully_connected_search:  # type: ignore
            self._model_choices.append(ParametersModel.select_fully_connected_model)

        if self.config.include_frame_stacking_search:  # type: ignore
            self._model_choices.append(ParametersModel.select_framestacking_model)

        if self.config.include_gtrxl_search:  # type: ignore
            self._model_choices.append(ParametersModel.select_gtrxl_model)

    @staticmethod
    def get_validator() -> type[HparamSearchValidator_AHBS]:
        """gets the configuration for AHBS

        Returns
        -------
        typing.Type[HparamSearchValidator_AHBS]
            validator
        """
        return HparamSearchValidator_AHBS

    def add_algorithm_hparams(self, rllib_config, tune_config) -> None:
        """[summary]

        Parameters
        ----------
        run_or_experiment_config : [type]
            [description]
        tune_config : [type]
            [description]
        args : [type]
            [description]
        """
        ahbs = tune.schedulers.AsyncHyperBandScheduler(
            time_attr=self.config.time_attr,  # type: ignore
            metric=self.config.metric,  # type: ignore
            mode=self.config.mode,  # type: ignore
            max_t=self.config.max_t,  # type: ignore
            grace_period=self.config.grace_period,  # type: ignore
            brackets=self.config.brackets,  # type: ignore
        )
        tune_config["num_samples"] = self.config.samples  # type: ignore
        tune_config["scheduler"] = ahbs
        rllib_config.update(ParametersPPO.ppo_hyperparameters())
        rllib_config["model"] = tune.sample_from(partial(ParametersModel.select_model, self._model_choices))



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Records an evaluation outcome to a folder
"""
import contextlib
import logging
import shutil
from datetime import datetime
from pathlib import Path
from shutil import copyfile

import pandas as pd
import ray.cloudpickle as pickle
from pydantic import BaseModel, field_validator

from corl.evaluation.evaluation_outcome import EvaluationOutcome
from corl.evaluation.recording.i_recorder import IRecord, IRecorder


class FolderRecord(IRecord):
    """A folder record which can save and load EvaluationOutcomes to folders"""

    _absolute_path: Path

    def __init__(self, absolute_path: Path | str):
        self._absolute_path = Path(absolute_path)
        self._logger = logging.getLogger(type(self).__name__)

    @property
    def absolute_path(self) -> Path:
        """Get the absolute path"""
        return self._absolute_path

    def save(self, outcome: EvaluationOutcome) -> None:
        """Save an evaluation outcome to a folder

        Arguments:
            outcome {EvaluationOutcome} -- Evaluation outcome to save
        """
        output_folder: Path = self._absolute_path

        for test_case_idx, episode_artifacts in outcome.episode_artifacts.items():
            # create test case folder
            test_case_folder = output_folder / f"test_case_{test_case_idx!s}"
            test_case_folder.mkdir(exist_ok=True)

            # Copy episode artifact files to test_case_folder
            for episode_artifact in episode_artifacts:
                start_time: str = episode_artifact.start_time.strftime("%Y-%m-%d_%H-%M-%S")

                for artifact_filename in episode_artifact.artifacts_filenames.values():
                    artifact_path = Path(artifact_filename)
                    new_path = test_case_folder / f"{start_time}_{artifact_path.name}"

                    try:
                        copyfile(artifact_path, new_path)
                    except FileNotFoundError as e:
                        self._logger.error(f"{e}")

                with open(test_case_folder / f"{start_time}_episode_artifact.pkl", "wb") as f:
                    # HACK todo fix
                    tmp_done_config = None
                    with contextlib.suppress(AttributeError):
                        tmp_done_config = episode_artifact.done_config
                        del episode_artifact.done_config

                    with contextlib.suppress(KeyError):
                        del episode_artifact.env_config["epp_registry"]

                    pickle.dump(episode_artifact, f)
                    if tmp_done_config is not None:
                        episode_artifact.done_config = tmp_done_config

            with open(test_case_folder / f"{start_time}_test_case.csv", "w") as f:
                outcome.get_test_cases().iloc[test_case_idx].to_csv(f)

        with open(output_folder / "test_cases.pkl", "wb") as f:
            pickle.dump(outcome.test_cases, f)

        with open(output_folder / "test_cases.csv", "w") as f:
            outcome.get_test_cases().to_csv(f)

    def load(self) -> EvaluationOutcome:
        """Load EvaluationOutcome from folder

        Returns:
            EvaluationOutcome -- Evaluation Outcome loaded from folder
        """

        with open(self._absolute_path / "test_cases.pkl", "rb") as f:
            test_cases = pickle.load(f)

        paths = self._absolute_path.glob("test_case_*/*_episode_artifact.pkl")
        artifacts = {}
        for path in paths:
            with open(path, "rb") as f:
                artifact = pickle.load(f)
                if artifact.test_case in artifacts:
                    raise RuntimeError(
                        f"Test case of {artifact.test_case} already exists. This indicates an issue with evaluation generation"
                    )
                artifacts[artifact.test_case] = [artifact]

        return EvaluationOutcome(test_cases, artifacts)


class Folder(BaseModel, IRecorder[FolderRecord]):
    """Generates a record to interface with a folder structure"""

    append_timestamp: bool = True
    dir: Path  # noqa: A003

    @field_validator("dir")
    def generate_output_dir(cls, v, info):
        """Generate the output dir"""

        if info.data["append_timestamp"]:
            timestamp_str = "_" + datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            output_folder = v.with_name(v.stem + timestamp_str)
        else:
            output_folder = v

        if output_folder.exists():
            shutil.rmtree(str(output_folder))

        # Make absolute
        output_folder = output_folder.resolve()
        output_folder.mkdir(parents=True, exist_ok=False)
        return output_folder

    def resolve(self) -> FolderRecord:
        """Resolve and generate the record

        Returns:
            FolderRecord -- Record
        """
        return FolderRecord(self.dir)


class EpisodeArtifactLoggingCallbackLoader(IRecord):
    """Loads files saved from enabling the EpisodeArtifactLoggingCallback during
    training"""

    def __init__(self, absolute_path: Path | str):
        """_summary_

        Parameters
        ----------
        absolute_path : typing.Union[Path, str]
            This is the absolute path to the
        """

        self._absolute_path = Path(absolute_path)

    @property
    def absolute_path(self) -> Path:
        """Get the absolute path to the root directory where the trajectories
        are stored. (e.g. <experiment_name>/trajectories/epoch_000001)
        """

        return self._absolute_path

    def save(self, _):
        """Overwrite the abstract class"""

    def load(self) -> EvaluationOutcome:
        episode_artifact_files = list(self._absolute_path.glob("*.p*kl*"))
        episode_artifact_files.sort()
        artifacts_dict = {}
        # The test_cases will store initial conditions for each episode
        # artifact
        test_cases = []
        for idx, epi_artifact_file in enumerate(episode_artifact_files):
            with open(epi_artifact_file, "rb") as file_obj:
                episode_artifact = pickle.load(file_obj)
                # The episode artifact test case number doesn't matter
                # for artifacts saved from training
            episode_artifact.test_case = idx
            artifacts_dict[idx] = episode_artifact

            # Build the test case dataframe: the test cases dataframe stores the reset parameters
            # we grab the information from the episode artifact
            sim_reset_dict = episode_artifact.params
            # Add prefix so the columns match with what gets generated during IterateTestCases
            sim_reset_dict = {f"environment.{k}": v for k, v in sim_reset_dict.items()}
            sim_reset_dict["test_case"] = idx
            test_cases.append(sim_reset_dict)
        if test_cases:
            test_cases_df = pd.DataFrame(test_cases)
            test_cases_df = test_cases_df.set_index("test_case")

            return EvaluationOutcome(test_cases=test_cases_df, episode_artifacts=artifacts_dict)
        return EvaluationOutcome(test_cases=test_cases, episode_artifacts=artifacts_dict)
